import collections
import math
from pathlib import Path
import tempfile
import os
import shutil
import sys
import warnings
import itertools
from uuid import uuid4
import hashlib

from digichem.datas import get_resource
import digichem.log

def checksum(*paths, hash_func = "sha1", buf_size = 1024 * 1024, ret_size = False):
    """
    Calculate the checksum of a file.

    This function avoids reading the entire file into memory at once.

    :param paths: The file(s) to calculate.
    :param hash_func: The name of a hashlib function to pass to haslib.new.
    :param buf_size: How much to read in a single pass. The default is 1MB.
    :param ret_size: If True, this function will return a tuple of (checksum, file_size). Otherwise, only the checksum is returned.
    """
    hasher = hashlib.new(hash_func)
    tot_size = 0

    for pth in paths:
        with open(pth, "rb") as file:
            while True:
                data = file.read(buf_size)

                if len(data) == 0:
                    # End of file
                    break

                tot_size += len(data)
                hasher.update(data)
    
    if ret_size:
        return (hasher.hexdigest(), tot_size)

    else:
        return hasher.hexdigest()

def expand_path(pth):
    """
    Expand variables (both $VAR and '~') in a string.
    
    This function is similar to calling both os.path.expanduser() and os.path.expandvar(),
    but with some additional functionality for relative paths (which are interpreted relative
    to the silico data dir).
    """
    pth = str(pth)
    
    new_pth = pth.replace("$SILICO", str(get_resource("data")))
    if new_pth != pth:
        warnings.warn("the '$SILICO' magic variable is deprecated, use '$DIGICHEM' instead", DeprecationWarning)

    pth = new_pth
    pth = pth.replace("$DIGICHEM", str(get_resource("data")))
    pth = os.path.expanduser(pth)
    pth = os.path.expandvars(pth)
    return pth

def tail(file, num_lines = 20):
    """
    Return the last n lines of a file.
    """
    last_lines = collections.deque(maxlen = num_lines)
    
    # We'll read through the file from the top.
    # This is probably inefficient for huge files but is easy to implement and I don't care just now.
    for line in file:
        last_lines.append(line)
        
    return list(last_lines)


def smkdir(dir_name, max_iter = math.inf):
    """
    Safe mkdir.
    
    Will attempt to create the given directory. If a file with the given name already exists, a directory with the same name except with an appropriate number appended will be created instead.
    
    If max_iter == 1, this method acts just like a normal mkdir().
    
    :param dir_name: The path to a directory to try and create.
    :param max_iter: The maximum number of iterations to try and create the directory. Default is no max.
    :returns: The path to the directory actually created.
    """
    # First try and make our base directory.
    counter = 1
    dir_name = str(dir_name)
    directory = None
    while True:
        try:
            directory = Path(dir_name + " {}".format(str(counter).zfill(2)) if counter != 1 else dir_name)
            directory.mkdir(parents = True)
            break
        except FileExistsError:
            if counter < max_iter:
                counter +=1
            else:
                raise
            
    return directory


# TODO: Would be nice to implement this as a context manager that acts like a file (better compatibility with yaml.dump() for example).
def atomic_write(file, data):
    """
    Atomically write to a file.
    
    :param file: The name of file to write to.
    :param data: Data to write to the file.
    """
    file = Path(file)
    # Open a temp file in the same directory as the real file (so we're likely to be on the same file system).
    with tempfile.NamedTemporaryFile("wt", dir = file.parent, delete = False) as temp_write:
        # Write data to the temp file.
        temp_write.write(data)
        # Force it to disk.
        temp_write.flush()
        os.fsync(temp_write)
        # Rename the temp file over the real file (overwriting it).
        os.rename(temp_write.name, file)


# Based on https://stackoverflow.com/questions/431684/how-do-i-change-the-working-directory-in-python/24176022#24176022
class cd:
    """
    Context manager for temporarily changing the working directory.
    """
    
    def __init__(self, directory):
        """
        Constructor for cd.
        
        :param directory: Path to the directory to change to.
        """
        self.new_directory = Path(directory)
        self.old_directory = None
        
    def __enter__(self):
        # Save our current working directory.
        self.old_directory = os.getcwd()
        
        # And change to the new directory.
        os.chdir(str(self.new_directory))
        
    def __exit__(self, exc_type, exc_value, traceback):
        # Restore our old directory.
        os.chdir(self.old_directory)
        

def copytree(src, dst, symlinks = False, ignore = None, copy_function = shutil.copy):
    """
    Fixed implementation of shutil.copytree that doesn't arbitrarily fail if src exists.
    
    Adapted from https://stackoverflow.com/questions/1868714/how-do-i-copy-an-entire-directory-of-files-into-an-existing-directory-using-pyth
    """
    if not os.path.exists(dst):
        os.makedirs(dst)
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            copytree(s, d, symlinks = symlinks, ignore = ignore, copy_function = copy_function)
        else:
            copy_function(s, d)
            
def copy_contents(src, dst, copy_function = shutil.copy):
    """
    Copy the top-level contents of one directory to another.
    """
    Path(dst).mkdir(parents = True, exist_ok = True)
    for child in Path(src).iterdir():
        if child.is_file():
            copy_function(child, Path(dst, child.name))
            
            
def rmtree(path, *args, **kwargs):
    """
    Fixed implementation of shutil.rmtree that doesn't arbitrarily require the first path to be a directory.
    """
    try:
        os.remove(path)
        
    except IsADirectoryError:
        shutil.rmtree(path, *args, **kwargs)


class Multi_file_wrapper():
    """
    A class that can be used to transparently handle both 'normal' files and stdin/stdout.
    """
    
    def __init__(self, file, mode = "r", *args, **kwargs):
        # If the filename is the special symbol '-' (a dash), then 'open' stdout as appropriate).
        if file == '-':
            if 'w' in mode or 'a' in mode:
                # We are writing.
                if 'b' in mode:
                    # We are a binary format, use a different stdout that accepts binary output.
                    object.__setattr__(self, 'file', sys.stdout.buffer)
                    #self.file = sys.stdout.buffer
                else:
                    object.__setattr__(self, 'file', sys.stdout)
                    #self.file = sys.stdout
            elif '+' in mode:
                # We are updating (not allowed).
                raise ValueError("Invalid mode specified '{}', updating is not valid for stdin/stdout".format(mode))
            else:
                # We are reading.
                if 'b' in mode:
                    object.__setattr__(self, 'file', sys.stdin.buffer)
                    #self.file = sys.stdin.buffer
                else:
                    object.__setattr__(self, 'file', sys.stdin)
                    #self.file = sys.stdin
            # We don't close stdin/stdout.
            object.__setattr__(self, 'should_close_file', False)
            #self.should_close_file = False
        else:
            # Normal file.
            
            object.__setattr__(self, 'file', open(file, mode, *args, **kwargs))
            object.__setattr__(self, 'should_close_file', True)
            #self.file = open(file, mode, *args, **kwargs)
            #self.should_close_file = True
            
    def __getattr__(self, name):
        """
        Magic method so we can pretend to be a real file.
        """
        return getattr(self.file, name)
    
    def __setattr__(self, name, value):
        """
        Magic method so we can pretend to be a real file.
        """
        return setattr(self.file, name, value)
        
    def __delattr__(self, name):
        """
        Magic method so we can pretend to be a real file.
        """
        if name in self.__dict__:
            object.__delattr__(self, name)
        else:
            return delattr(self.file, name)
            
            
    def close(self):
        if self.should_close_file:
            self.file.close()
        else:
            self.file.flush()
            
    def __enter__(self):
        """
        Magic function for the 'with' keyword.
        """
        return self
        
    def __exit__(self, etype, value, traceback):
        """
        Magic function for the 'with' keyword, called at the end of the block.
        """
        # Close our file.
        self.close()

class Safe_path():
    """
    Get a 'safe' path to a file.

    A safe path is made up of only alphanumeric characters, and is short. This function is useful for dealing
    with antiquated programs that struggle with whitespace/non-alpha characters or have a max file name requirement.

    This class should be used as a context manager. The returned path is a temporary symbolic link link to the true file.
    The symbolic link will be created in the specified 'dir', which defaults to the CWD.
    """

    def __init__(self, unsafe_path, dir = "./", suffix = ""):
        self.unsafe_path = unsafe_path
        self.dir = dir
        self.link = None
        self.suffix = suffix


    def enter(self):
        """
        Create the symlink.
        """
        self.link = Path(self.dir, ".{}".format(uuid4().hex) + self.suffix)
        os.symlink(self.unsafe_path, self.link)

    def close(self):
        """
        Remove the symlink.
        """
        try:
            os.unlink(self.link)
        
        except FileExistsError:
            digichem.log.get_logger().warning("Failed to remove temporary symbolic link '{}' -> '{}'; the file has already disappeared?".format(self.link, self.unsafe_path))
            
    def __enter__(self):
        """
        Magic function for the 'with' keyword.
        """
        self.enter()
        return self
        
    def __exit__(self, etype, value, traceback):
        """
        Magic function for the 'with' keyword, called at the end of the block.
        """
        # Close our file.
        self.close()

def dir_size(target):
    """
    Calculate the total used file space of a directory and all contents.
    """
    bytes = 0
    for path in itertools.chain(Path(target).rglob("*"), [Path(target)]):
        bytes += path.stat().st_size
    
    return bytes


import math
import numpy
import itertools
import scipy.constants
import scipy.signal

from digichem.misc.base import transpose
from digichem.exception.base import Digichem_exception
import digichem.result.excited_state

class Spectroscopy_graph_abc():
    """
    ABC for classes used to plot Gaussians over spectroscopic results. 
    """
    
    @property
    def coordinates(self):
        """
        Get the coordinates around which this graph is plotted.
        
        Note that this property may be transformed into different units, use base_coordinates for untransformed variant.
        """
        raise NotImplementedError("Implement in subclass")
    
    def peaks(self):
        """
        Find peaks in the cumulative graph.
        
        :return: A list of x,y coords that are peaks.
        """
        coords = self.plot_cumulative_gaussian()
        y_coords = [y for x,y in coords]
        indexes = scipy.signal.find_peaks(y_coords, height = max(y_coords) * self.cutoff)[0]
        return [coords[index] for index in indexes]
        
    def plot_gaussian(self):
        """
        Plot a gaussian distribution around a set of coordinates.
        :return: A list of lists of tuples of (x, y) coordinates plotted by the gaussian function (one list per input coordinate).
        """
        raise NotImplementedError("Implement in subclass")
    
    def plot_cumulative_gaussian(self):
        """
        Plot an additive gaussian distribution around a set of coordinates.
        
        :return: A single list of tuples of (x, y) coordinates plotted by the gaussian function.
        """
        # First, get our gaussian plots.
        gplots = self.plot_gaussian()
        
        # Sum our y values.
        coords = {}
        for plot in gplots:
            for x, y in plot:
                coords[x] = coords.get(x, 0) + y
                
        # Now return as list of tuples.
        return sorted(list(coords.items()), key = lambda coord: coord[0])
    
    @classmethod
    def gaussian(self, a, b, c, x):
        """
        An implementation of the gaussian function.
        
        :param a: The maximum height of the peak.
        :param b: The x position of the center of the peak.
        :param c: The width of the peak.
        :param x: The x-value to plot for.
        :return: The corresponding y value.
        """
        return a * math.exp(-( (x - b)**2 / (2 * c**2) ))
    
    @classmethod
    def gaussian_x(self, a, b, c, y):
        """
        An implementation of the gaussian function rearranged to find x.
        
        :param a: The maximum height of the peak.
        :param b: The x position of the center of the peak.
        :param c: The width of the peak.
        :param y: The y-value to plot for.
        :return: A tuple of the two corresponding x values.
        """
        # We have two solutions of the form x = ± d + b
        # Calculate d first.
        try:
            d = math.sqrt( -math.log( y/a ) * 2 * c **2  )
        except ZeroDivisionError:
            # a (max height) is zero; the intensity is zero.
            # We could instead return b?
            raise Digichem_exception("'{}' cannot calculate Gaussian function limits; a (intensity) is zero".format(self.__name__))
        
        # Now return our two solutions.
        return (-d + b, d + b)
    
    @classmethod
    def fwhm_to_c(self, fwhm):
        """
        Convert a full width at half maximum to the corresponding c-value in the Gaussian function.
        
        :param fwhm: The desired full width at half maximum (in nm).
        :return: The c-value.
        """
        return fwhm / (2 * math.sqrt(2 * math.log(2)))


class Spectroscopy_graph(Spectroscopy_graph_abc):
    """
    Top level class for graphing spectroscopy results (energy vs intensity).
    
    For generating pictures of these graphs, see digichem.image.spectroscopy
    """
    
    def __init__(self, coordinates, fwhm, resolution = 1, cutoff = 0.01, filter = 1e-6, adjust_zero = False):
        """
        Constructor for Spectroscopy_graph objects
        
        :param coordinates: A list of (energy, intensity) tuples to plot. The units of energy and intensity are irrelevant here (but should be consistent). Note that coordinates with 0 intensity will be removed.
        :param fwhm: The full-width at half-maximum to plot peaks with (in units of the x axis).
        :param resolution: The spacing (or step-size) between points to plot using the gaussian function, in units of the x-axis.
        :param cutoff: The minimum y value to plot using the gaussian function, as a fraction of the intensity.
        :param adjust_zero: If True and all y values are 0, set all y values to 1 (so that something can be plotted).
        """
        # We save our coordinates under two properties.
        # base_coordinates are untransformed.
        # coordinates (which is by default the same as base_coordinates) are transformed.        
        self.false_intensity = False
        # Set arb y value if we've been asked to.
        if adjust_zero:
            # Check if all the y axis is zero.
            if all([y == 0 for x,y in coordinates]):
                coordinates = [(x, 1) for x, y in coordinates]
                self.false_intensity = True
        
        self.base_coordinates = [(x, y) for x,y in coordinates if y != 0]
        self.fwhm = fwhm
        self.resolution = resolution
        self.cutoff = cutoff
        self.filter = filter
        
        # Caches
        self._gaussians = []
        self._cumulative_gaussians = []
        
    @classmethod
    def from_vibrations(self, vibrations, *args, **kwargs):
        """
        Alternative constructor from a Vibrations_list object.
        """
        return self([(vibration.frequency, vibration.intensity) for vibration in vibrations], *args, **kwargs)
    
    @classmethod
    def from_excited_states(self, excited_states, *args, **kwargs):
        """
        An alternative constructor that takes a list of excited states as argument.
        
        :param excited_states: An Excited_states_list object to construct from.
        :param adjust_zero: If all the intensities of the given excited states are zero, whether to arbitrarily set the y coords to 1.
        :param **kwargs: Passed to the real constructor.
        """
        coords = [(excited_state.energy, excited_state.oscillator_strength if excited_state.oscillator_strength is not None else 0) for excited_state in excited_states]
        return self(coords, *args, **kwargs)
    
    @property
    def coordinates(self):
        """
        Get the coordinates around which this graph is plotted.
        
        Note that this property may be transformed into different units, use base_coordinates for untransformed variant.
        """
        return self.base_coordinates
        
    def plot_gaussian(self, refresh = False):
        """
        Plot a gaussian distribution around a set of coordinates.
        :return: A list of lists of tuples of (x, y) coordinates plotted by the gaussian function (one list per input coordinate).
        """
        if len(self._gaussians) == 0 or refresh:
            self._gaussians = self.get_gaussian()
        
        return self._gaussians
        
    def get_gaussian(self):
        # First, determine our c value.
        c = self.fwhm_to_c(self.fwhm)
        
        # Next, determine the limits in which we'll plot.
        limits = self.gaussian_limits()
        
        # Apply rounding to nearest resolution step to cover up floating-point errors.
        points = [self.resolution *round(point / self.resolution) for point in numpy.linspace(*limits)]
        
        # Plot and return.
        digichem.log.get_logger().debug(
            "Plotting gaussian peaks from {:0.2f} to {:0.2f} with a step size of {} ({} total points) for {} peaks ({} total iterations)".format(
                limits[0],
                limits[1],
                self.resolution,
                limits[2],
                len(self.base_coordinates),
                len(self.base_coordinates) * limits[2]
            )
        )
        gaussians = [
            [(x, y) for x in points if (y  := self.gaussian(a, b, c, x)) >= self.filter]
            for b, a in self.base_coordinates
        ]
        
        return gaussians
    
    def plot_cumulative_gaussian(self, refresh = False):
        """
        Plot an additive gaussian distribution around a set of coordinates.
        
        :return: A single list of tuples of (x, y) coordinates plotted by the gaussian function.
        """
        if len(self._cumulative_gaussians) == 0 or refresh:
            self._cumulative_gaussians = self.get_cumulative_gaussian()
        
        return self._cumulative_gaussians
        
    def get_cumulative_gaussian(self):
        # First, determine our c value.
        c = self.fwhm_to_c(self.fwhm)
        
        # Next, determine the limits in which we'll plot.
        limits = self.gaussian_limits()
        
        # Apply rounding to nearest resolution step cover up floating-point errors.
        points = [self.resolution *round(point / self.resolution) for point in numpy.linspace(*limits)]
        
        # Plot and return.
        digichem.log.get_logger().debug("Plotting cumulative gaussian peaks from {:0.2f} to {:0.2f} with a step size of {} ({} total points) for {} peaks ({} total iterations)".format(limits[0], limits[1], self.resolution, limits[2], len(self.base_coordinates), len(self.base_coordinates) * limits[2]))
        gaussian = [
            (x, y) for x in points if ( y := sum((self.gaussian(a, b, c, x) for b, a in self.base_coordinates))) > self.filter
        ]
        
        return gaussian
    
    def gaussian_limits(self):
        """
        Determine min and max x limits to plot a gaussian function.
        
        :param c: The width of the peak.
        :param cutoff: The minimum y value to plot using the gaussian function, as the fraction of a.
        :param resolution: The spacing between points to plot using the gaussian function, in units of the x-axis.
        :return: A tuple of (minlim, maxlim, num) where minlim is the most negative value, maxlim the most positive and num the integer number of points to plot between them to achieve resolution.
        """
        # Calculate limits for each set of coordinates given to us.
        all_limits = [self.gaussian_x(y, x, self.fwhm_to_c(self.fwhm), self.cutoff * y) for x, y in self.base_coordinates]
        
        try:
            limits = (min(itertools.chain.from_iterable(all_limits)), max(itertools.chain.from_iterable(all_limits)))
        
        except ValueError:
            if len(list(itertools.chain.from_iterable(all_limits))) == 0:
                # Nothing to plot.
                raise Digichem_exception("'{}' cannot plot spectrum; there are no values".format(type(self).__name__))
            else:
                raise
            
        # Align our start and end points to an integer multiple of resolution.
        # This is necessary to facilitate easy summation of gaussians (because they all align to the same grid,
        # assuming the same value of resolution)
        start = math.floor(limits[0] / self.resolution) * self.resolution
        end = math.ceil(limits[1] / self.resolution) * self.resolution
        
        # Now we need to generate the x values which we'll plot for.
        # This is a little more complicated than it needs to be because there's no simple range() for floats.
        # Calculate the number of points (and round up).
        num_points = round( math.fabs(end - start) / self.resolution)
        
        # Extend our limits so they are a clean multiple of num_points.
        limits = (start, start + num_points * self.resolution)

        return (limits[0], limits[1], num_points +1)
    
    
class Absorption_emission_graph(Spectroscopy_graph):
    """
    Class for graphing absorption/emission spectra in nm.
    
    Use the normal Spectroscopy_graph object for plotting in eV.
    """
    
    def __init__(self, *args, use_jacobian = True, **kwargs):
        """
        Constructor for Absorption_emission_graph objects.
        
        :param false_intensity: A flag to indicate that the intensity units (y-axis) are arbitrary.
        :param use_jacobian: Whether to use the jacobian transform to scale the y axis.
        """
        # Get our x,y values from our excited states.
        super().__init__(*args, **kwargs)
        self.use_jacobian = use_jacobian
        
    @property
    def coordinates(self):
        """
        Get the coordinates around which this graph is plotted.
        
        Note that this property may be transformed into different units, use base_coordinates for untransformed variant.
        """
        return self.nm_coordinates
    
    @property
    def nm_coordinates(self):
        """
        Get the list of coordinates of this graph scaled to nm.
        """
        return [self.energy_to_wavelength(coord, self.use_jacobian) for coord in self.base_coordinates]
    
    @classmethod
    def energy_to_wavelength(self, coord, use_jacobian):
        """
        Convert a pair of x,y coordinates in ev to nm.
        
        :param coord: A tuple of (eV, f).
        :return: A tuple of (nm, i), where i is intensity.
        """
        x, y = coord
        return (digichem.result.excited_state.Excited_state.energy_to_wavelength(x), self.jacobian(x, y) if use_jacobian else y)
        
    @classmethod
    def jacobian(self, E, f_E):
        """
        An implementation of the jacobian transform that scales intensity in energy units to intensity in wavelength units.
        
        See J. Phys. Chem. Lett. 2014, 5, 20, 3497 for why this is necessary.
        
        Note that the jacobian transform will maintain the area under the curve regardless of x units (nm or x).
        Sadly, this has the consequence of mangling the intensity units (it becomes tiny; an oscillator strength of 1 at 3 eV becomes 1.163e-12).
        """
        return ((E * scipy.constants.electron_volt)**2  * f_E) / (scipy.constants.Planck * scipy.constants.c)
    
    @classmethod
    def inverse_jacobian(self, E, f_nm):
        """
        An implementation of the jacobian transform that scales intensity in wavelength units to intensity in energy units.
        
        See J. Phys. Chem. Lett. 2014, 5, 20, 3497 for why this is necessary.
        
        Note that the jacobian transform will maintain the area under the curve regardless of x units (nm or x).
        Sadly, this has the consequence of mangling the intensity units (it becomes tiny; an oscillator strength of 1 at 3 eV becomes 1.163e-12).
        """
        # TODO: Might be better to rearrange this to accept nm rather than eV?
        return (
            (f_nm * scipy.constants.Planck * scipy.constants.c) / (E * scipy.constants.electron_volt) **2
        )
    
    @classmethod
    def shift_coord(self, coord, delta_eV):
        """
        Shift a coordinate (in nm) by a given energy value.

        :param delta_eV: The energy (in eV) to shift by. A positive value will blueshift (higher energy).
        """
        old_x_nm, old_y_nm = coord
        # Convert x to energy.
        old_x_ev = digichem.result.excited_state.Excited_state.wavelength_to_energy(old_x_nm)
        # Transform y.
        old_y_ev = self.inverse_jacobian(old_x_ev, old_y_nm)

        # Shift by given amount.
        new_x_ev = old_x_ev + delta_eV

        # Convert back to nm.
        new_x_nm, new_f_nm = self.energy_to_wavelength((new_x_ev, old_y_ev), True)

        return (new_x_nm, new_f_nm)
    
    def shift(self, delta_eV):
        """
        """
        return map(lambda coord: self.shift_coord(coord, delta_eV), self.coordinates)

    
    def plot_gaussian(self):
        """
        Plot a gaussian distribution around our excited state energies.
        
        :return: A list of lists of tuples of (x, y) coordinates plotted by the gaussian function (one list per input coordinate).
        """
        # All we need to do over our parent is convert x values from e to wavelength.
        # And scale y values using the jacobian transform.
        return [[self.energy_to_wavelength(coord, self.use_jacobian) for coord in plot] for plot in super().plot_gaussian()]
    
    def plot_cumulative_gaussian(self):
        """
        Plot an additive gaussian distribution around a set of coordinates.
        
        :return: A single list of tuples of (x, y) coordinates plotted by the gaussian function.
        """
        return [self.energy_to_wavelength(coord, self.use_jacobian) for coord in super().plot_cumulative_gaussian()]


# TODO: Find this a better home?
def unpack_coupling(couplings, satellite_threshold = 0.02):
    """
    Unpack a nested dict of dict of NMR_group_spin_coupling objects into a single, ordered dict.
    
    The keys of the returned dict will be a tuple of (foreign_atom_group, foreign_isotope).
    """
    couplings = {
            (coupling_group, coupling_isotope): isotope_coupling for coupling_group, atom_dict in couplings.items() for coupling_isotope, isotope_coupling in atom_dict.items()
            if coupling_group.element[coupling_isotope].abundance / 100 > satellite_threshold
        }
    # Sort couplings.
    couplings = dict(sorted(couplings.items(), key = lambda coupling: abs(coupling[1].total), reverse = True))
    
    return couplings


class NMR_graph(Spectroscopy_graph):
    """
    A class for plotting a single NMR peak.
    
    A 'single' NMR peak here refers to the signal that would be observed from only one atom (or atom group).
    For plotting an entire spectrum, use a Combined_graph of NMR_graph objects.
    """
    
    def multiplicity(self, atom_group, coupling, satellite_threshold = 0.02):
        """
        Determine the multiplicity of this peak.
        
        Note that the returned multiplicity might not match exactly the observed multiplicity of the peak
        due to line-broadening and overlapping signals.
        
        :param atom_group: The atom_group that this peak corresponds to.
        :param coupling: NMR_group_spin_coupling objects between this atom_group and other groups.
        """        
        # First, determine how many peaks are visible.
        peaks = self.peaks()
        
        # Filter out peaks which are significantly smaller than the tallest peak.
        max_peak = max(transpose(peaks, 2)[1])
        peaks = [peak for peak in peaks if peak[1] > max_peak * satellite_threshold]
        
        if len(peaks) == 1:
            return [{"symbol": "s", "number": 1, "multiplicity": "singlet"}]
        
        # A list of multiplicity dicts.
        mults = []
        total_peaks = 1
        coupling_index = 0
        
        # Get couplings.
        couplings = unpack_coupling(coupling)
        couplings = list(couplings.values())
        
        # Unless coupling has been calculated for all available isotopes for each atom group (unlikely),
        # there will be one additional peak for each atom group from non-NMR active nuclei.
        # Make sure we don't count this peak multiple times.
        #
        # For atoms in which the majority of the abundance is already accounted for (1H, for example),
        # exclude the residual peak.
        residual_isotope_peaks = set(
            [atom_group for atom_group, isotopes in coupling.items() if sum((atom_group.element[isotope].abundance for isotope in isotopes.keys())) / 100 > satellite_threshold]
        )
        
        # We will keep requesting more splitting until we are able to account for all the peaks we can see.
        while len(peaks) > total_peaks:
            # Get the next coupling available.
            try:
                group_coupling = couplings[coupling_index]
            
            except IndexError:
                # Ran out of coupling.
                break
            
            other_group = group_coupling.groups[group_coupling.other(atom_group)]
            residual = other_group not in residual_isotope_peaks
            multiplicity = group_coupling.multiplicity(atom_group)
            residual_isotope_peaks.add(other_group)
            
            mults.append(multiplicity)
            num = multiplicity['number']
            if residual:
                num += 1
            
            total_peaks *= num
            
            # Update counter.
            coupling_index += 1
        
        return mults
        
    @classmethod
    def from_nmr(self, nmr_peaks, *args, **kwargs):
        """
        An alternative constructor that takes a list of simulated NMR peaks as argument.
        
        :param nmr_peaks: Simulated NMR peaks. See result.nmr.NMR_spectrometer for how to obtain this data.
        """
        return self(nmr_peaks, *args, **kwargs)


class Combined_graph(Spectroscopy_graph_abc):
    """
    A class for plotting one graph from multiple, individual graphs.
    """
    
    def __init__(self, graphs):
        """
        Constructor for Combined_graph objects.
        
        :param graphs: A dictionary of individual Spectroscopy_graph objects to combine.
        """
        self.graphs = graphs
        
    @property
    def cutoff(self):
        """
        """
        return max((graph.cutoff for graph in self.graphs.values()))
    
    @property
    def fwhm(self):
        """
        """
        return sum((graph.fwhm for graph in self.graphs.values())) / len(self.graphs)
        
    @classmethod
    def from_nmr(self, grouped_peaks, *args, **kwargs):
        """
        Construct a Combined_graph from a dictionary of grouped NMR peaks
        
        :param grouped_peaks: A dictionary of lists of NMR peaks. Each key in the dictionary should correspond to one atom group.
        """
        return self(
            {
                peak_key: NMR_graph.from_nmr(peaks, *args, **kwargs) for peak_key, peaks in grouped_peaks.items()
            }
        )
    
    @property
    def coordinates(self):
        """
        Get the coordinates around which this graph is plotted.
        
        Note that this property may be transformed into different units, use base_coordinates for untransformed variant.
        """
        return list(itertools.chain(*[graph.coordinates for graph in self.graphs.values()]))
        
    def plot_gaussian(self):
        """
        Plot a gaussian distribution around a set of coordinates.
        :return: A list of lists of tuples of (x, y) coordinates plotted by the gaussian function (one list per input coordinate).
        """
        # Because this class represents one additional layer of abstractions,
        # we'll combine the results returned by each sub graph together, to obtain a single nested list.
        return [
            graph.plot_cumulative_gaussian()
            for graph in self.graphs.values()
        ]
    


from digichem.parse.cclib import Cclib_parser
import digichem.file.types as file_types
import digichem.log


class Orca_parser(Cclib_parser):
    """
    Top level class for parsing output from Gaussian log files.
    """
    
    # A dictionary of recognised auxiliary file types.
    INPUT_FILE_TYPES = {
            file_types.orca_gbw_file: "gbw_file",
            file_types.orca_density_file: "density_file",
            file_types.orca_density_info_file: "density_info_file"
        }
    
    def parse_output_line(self, log_file, line):
        """
        Perform custom line-by-line parsing of an output file.
        """
        
        # Spin-orbit coupling.
        # We're looking to populate two or three attributes
        # 'socstates'   : A two-membered list of the singlet and triplet symbols that make up this coupling (eg, ["S(1)", "T(2)"].
        # 'socenergies' : The total spin-orbit coupling value (RSS of socelements).
        # 'socelements' : A three-membered list of the soc values for the triplet state with number +1, 0 and -1.
        if "CALCULATED SOCME BETWEEN TRIPLETS AND SINGLETS" in line:
            # Start of the SOC section.
            # The same header is used for SOC in cartesian basis (x,y,z) and for individual triplet states (+1, 0, -1).
            line = next(log_file)
            line = next(log_file)
            line = next(log_file)
            
            soc_type = None
            if "Z" in line and "X" in line and "Y in line":
                # Cartesian SOC
                # In this format we can parse total SOC only.
                soc_type = "cartesian"
            elif "0" in line and "-1" in line and "+1" in line:
                # Triplet SOC.
                # In this format we can parse individual SOC as well as total SOC.
                soc_type = "triplet"
            
            else:
                pass
                digichem.log.get_logger().debug("Unrecognised SOC section started by line '{}'".format(line))
            
            if soc_type is not None:
                # Reset our attributes.
                self.data.socstates = []
                self.data.socenergies = []
                if soc_type == "triplet":
                    self.data.socelements = []
                elif hasattr(self.data, 'socelements'):
                    delattr(self.data, 'socelements')
                
                line = next(log_file)
                line = next(log_file)
                
                while line.strip() != "--------------------------------------------------------------------------------" and \
                    line.strip() != "":
                    # Each line is the coupling between one singlet state and one triplet state.
                    # 1      1    (   0.00 ,    0.00)    (  -0.00 ,   -0.00)    (  -0.00 ,    0.00)
                    split_line = line.split()
                    try:
                        triplet_index = int(split_line[0])
                        singlet_index = int(split_line[1])
                    
                    except Exception:
                        print(line)
                    
                    # Split on brackets to get each xyz/0, -1, +1 element.
                    soc_elements = []
                    soc_element_strings = line.split("(")[1:]
                    
                    for soc_element_string in soc_element_strings:
                        # The last character will be the closing bracket, so we can discard.
                        real, imagine = [float(ele) for ele in soc_element_string.strip()[:-1].split(",")]
                        
                        # We're not interested in the real or imaginary parts, just combine (root of the sum of the squares).
                        soc_elements.append((real**2 + imagine**2)**0.5)
                    
                    # We now have everything we need.
                    self.data.socstates.append(["S({})".format(singlet_index), "T({})".format(triplet_index)])
                    self.data.socenergies.append((soc_elements[0]**2 + soc_elements[1]**2 + soc_elements[2]**2)**0.5)
                    
                    # Only add elements if they are triplet elements (not cartesian).
                    if soc_type == "triplet":
                        # The order of triplet states is different in Orca to what we expect.
                        # We want +1, 0, -1
                        # Orca is 0, -1, +1
                        self.data.socelements.append([soc_elements[2], soc_elements[0], soc_elements[1]])
                
                    line = next(log_file)
                
                


from itertools import chain, combinations

def powerset(iterable):
    "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))

def regular_range(median, number, spacing):
    """
    Generate a range of numbers with a given median and spacing.
    """
    if number % 2 == 0:
        # Even number of peaks
        peaks = []
        for magnitude in range(int(number /2)):
            peaks.append(median + (spacing /2 + magnitude * spacing))
            peaks.append(median - (spacing /2 + magnitude * spacing))
        
    else:
        # Odd number of peaks.
        peaks = [median]
        
        for magnitude in range(int((number -1) /2)):
            magnitude = magnitude+1
            peaks.append(median + magnitude * spacing)
            peaks.append(median - magnitude * spacing)
        
    return sorted(peaks)
    

def dict_list_index(dictionary, item):
    "Find the key in a dictionary which contains the list which contains an item."
    for dict_key, dict_value in dictionary.items():
        if item in dict_value:
            return dict_key
    
def dict_get(dict_obj, *fields):
    """
    Get a value from a nested dict of arbitrary depth.
    
    :param dict_obj: The nested dict.
    :param fileds: Names of nested keys.
    """
    if len(fields) == 1:
        return dict_obj[fields[0]]
    else:    
        return dict_get(dict_obj[fields[0]], *fields[1:])
    
def transpose(nested_list, dimensions):
    """
    Transpose a nested list, returning separate lists of the nested items.
    
    eg, transpose([(1, "a"), (2, "b"), (3, "c")])
    -> [[1, 2, 3], ["a", "b", "c"] 
    """
    if len(nested_list) == 0:
        return [list() for i in range(dimensions)]
    
    else:
        return list(map(list, zip(*nested_list)))
    


# Methodologies for determining molecule linearity.
import math

from statistics import pstdev
from digichem.result.alignment import Alignment

class Furthest_atom_pair(Alignment):
    """
    The 'kebab' (skewer) method for estimating molecule linearity. 
    """
        
    # Names that uniquely describe this alignment protocol.
    CLASS_HANDLE = ["Furthest Atom Pair", "FAP", "Kebab"]
        
    def align_axes(self):
        """
        Realign the axes of our coordinate system, so they have the following meaning:
        
        X-axis: The long axis (the kebab skewer), which we define as passing through the pair of atoms with the greatest separation in our set.
        Y-axis: The middle axis, which we define as perpendicular to the the X-axis (obviously) and passing through the atom that is furthest from the X-axis. Note that the Y-axis only has to pass through one atom; there may not be a corresponding atom on the other side (but there will be if the molecule is symmetrical about the X-axis).
        Z-axis: The short axis, defined as perpendicular to both the X and Y-axes (so we have no choice where this goes).
        
        :return: Nothing. The atoms are rearranged in place.
        """
        # First align our X axis.
        self.align_X()
        
        # Next we want to orientate our secondary (Y) axis.
        self.align_Y()
        
        # This does nothing (but might not in the future).
        self.align_Z()
        
    def align_X(self):
        """
        Align the X axis.
        
        You do not need to call this method yourself; it is called as part of align_axes().
        """
        # First get our most separated atoms. This is probably inefficient.
        # This is a tuple of (distance, atom1, atom2).
        furthest = self.get_furthest_atom_pair()
                    
        # Our new origin is halfway along these two points.
        new_origin = ( (furthest[1].coords[0] + furthest[2].coords[0]) /2, (furthest[1].coords[1] + furthest[2].coords[1]) /2, (furthest[1].coords[2] + furthest[2].coords[2]) /2)
        # Translate to new origin.
        self.translate((-new_origin[0], -new_origin[1], -new_origin[2]))
        
        # Now we need to rotate, because we're going to set these two atoms as points on the x axis.
        # We rotate twice, once to set y = 0, once to set z = 0.
        
        # Rotate xy coords (along Z axis).
        # First determine angle (we should be able to use either atom as reference because we rotate about their midpoint).
        theta = self.get_theta(furthest[1].coords[1], furthest[1].coords[0])
        self.rotate_XY(theta)
        
        # Rotate xz coords (along Y axis).
        # Get our angle again.
        theta = self.get_theta(furthest[1].coords[2], furthest[1].coords[0])
        self.rotate_XZ(theta)
        
    def align_Y(self):
        """
        Align the Y axis.
        
        You do not need to call this method yourself; it is called as part of align_axes().
        """
        # Note that we're not just looking for the atom furthest away, we're looking for the greatest distance across.
        
        # Get the most separated pair of atoms in this plane.
        furthest = self.get_furthest_atom_pair_in_YZ()
            
        # Work out the angle between the two atoms.
        theta = self.get_theta(furthest[1].coords[2] - furthest[2].coords[2], furthest[1].coords[1] - furthest[2].coords[1])
        
        # Rotate so the Y axis is parallel to a line drawn between our two furthest atoms in this plane.
        self.rotate_YZ(theta)
        
    def align_Z(self):
        """
        Align the Z axis.
        
        You do not need to call this method yourself; it is called as part of align_axes().
        """
        # We don't need to align our Z axis, as it is automatically defined once we've aligned out X and Y.
        pass
    
    def get_deviation_from_X(self):
        # First get the distance of each atom from the X-axis.
        distance = [math.sqrt( (atom.coords[1])**2 + (atom.coords[2])**2 ) for atom in self]
        # Get the deviation.
        return pstdev(distance)
                
    def get_furthest_atom_pair(self):
        """
        Get the pair of atoms in our set that are separated by the greatest distance.
        
        :return: A tuple of the form (distance, atom1, atom2) where 'distance' is the straight line distance between 'atom1' and 'atom2'. The order of 'atom1' vs 'atom2' is essentially random and is irrelevant.
        """
        # This is a tuple of (distance, atom1, atom2).
        furthest = (0, None, None)
        
        for atom in self:
            for foreign_atom in self:
                # Compute distance.
                distance = (atom.distance(foreign_atom), atom, foreign_atom)
                # Check to see if we're further.
                if distance[0] >= furthest[0]:
                    # This distance is greater, so update.
                    furthest = distance
        
        # And return.
        return furthest
    
    def get_furthest_atom_from_X(self):
        """
        Get the atom that is furthest from the X-axis.
        
        :return: A tuple of the form (distance, atom) where 'distance' is the straight line distance from 'atom' to the X-axis.
        """
        furthest = (0, None)
        for atom in self:
            # Calculate distance
            distance = math.sqrt((atom.coords[1])**2 + (atom.coords[2])**2)
            # Check.
            if distance >= furthest[0]:
                # Set.
                furthest = (distance, atom)
                
        # And return
        return furthest
    
    def get_furthest_atom_pair_in_YZ(self):
        """
        Get the pair of atoms in our set that are separated by the greatest distance in the YZ plane.
        
        :return: A tuple of the form (distance, atom1, atom2) where 'distance' is the  distance between 'atom1' and 'atom2'. The order of 'atom1' vs 'atom2' is essentially random and is irrelevant.
        """
        furthest = (0, None, None)
        for atom in self:
            for foreign_atom in self:
                # Get the greatest distance in this plane.
                distance = (math.sqrt( (atom.coords[1] - foreign_atom.coords[1])**2 + (atom.coords[2] - foreign_atom.coords[2])**2 ), atom, foreign_atom)
                if distance[0] >= furthest[0]:
                    furthest = distance
        # And return.
        return furthest
        


from matplotlib import ticker
from matplotlib.ticker import FuncFormatter
from itertools import chain
import math

from digichem.image.graph import OneD_graph_image_maker
#from digichem.result.excited_state import Excited_state, Excited_state_list
import digichem.result.excited_state

class Excited_states_diagram_maker(OneD_graph_image_maker):
    """
    Class for making orbital energy diagrams.
    """
    
    def __init__(self, output, excited_states, ground_state, y_limits = "auto", show_dest = True, **kwargs):
        super().__init__(output, **kwargs)
        self.excited_states = excited_states
        self.ground_state = ground_state
                        
        # Each multiplicity will be plotted as a separate column, so first we need to organise our data.
        states = chain(self.excited_states, [self.ground_state]) if self.ground_state is not None else self.excited_states
        self.grouped_states = (type(self.excited_states)(sorted(states, key = lambda state: state.energy))).group()
        # We'll also keep hold of just our excited states (so no ground).
        self.grouped_excited_states = self.excited_states.group()
        
        # Whether to display our dEst label, this is forced off if we don't have only singlets and triplets.
        self.show_dest = show_dest if 1 in self.grouped_excited_states and 3 in self.grouped_excited_states and len(self.grouped_excited_states) == 2 else False
        self.left_padding = 0
        self.right_padding = 0    
            
        # Set our output dpi (higher values = bigger image/quality).
        self.output_dpi = 140
        
        # Initial image dimensions (we'll change these to suit).
        self.init_image_width = 5.6
        self.init_image_height = 5.6
        
        # Amount of space to allocate per unit of each axis.
        self.inch_per_x = 1.66
        #self.inch_per_y = 0.94
        self.inch_per_y = 1.1
        
        # Y axis label (we don't have an x axis label).
        self.y_label = "Energy /eV"
        
        self.plot_options = {
            "orientation": "vertical",
            "lineoffsets": 1,
            "linelengths": 0.9,
            "linewidths": None,
            "colors": None
        }
        
        # Save our scaling method.
        self.limits_method = y_limits
    
    @classmethod
    def from_options(self, output, *, excited_states, ground_state, options, **kwargs):
        """
        Constructor that takes a dictionary of config like options.
        """    
        return self(
            output,
            excited_states = excited_states,
            ground_state = ground_state,
            show_dest = options['excited_states_diagram']['show_dest'],
            y_limits = options['excited_states_diagram']['y_limits'],
            enable_rendering = options['excited_states_diagram']['enable_rendering'],
            **kwargs
        )
            
    def plot_lines(self):
        """
        Plot the lines that make up the body of the graph.
        
        This method is called automatically as part of the make() method.
        :return: Nothing.
        """
        # Turn our dictionary of grouped excited states into an ordered list for matplotlib.
        data = [
            [excited_state.energy for excited_state in self.grouped_states[grouped_key]]
            #self.grouped_states[grouped_key] 
            for grouped_key 
            in sorted(self.grouped_states)
        ]
            
        # We can use an event plot which is good for 1D graphs like this.
        self.axes.eventplot(data, linestyles = "solid", **self.plot_options)
    
    def get_x_pos_from_multiplicity(self, multiplicity):
        """
        Get the X coordinate of an energy state.
        
        :raises ValueError: If the given multiplicity is not plotted. 
        :param multiplicity: The multiplicity of the energy state, which determines its x position.
        :return: The X coord.
        """
        # Get all the multiplicities that we're using.
        mult = sorted(self.grouped_states)
        
        # Our X position depends on the multiplicity of our GS and how many total multiplicities there are (as well as how spaced out each column on the graph is).
        if len(self.grouped_states) > 1:
            return mult.index(multiplicity) * self.plot_options['lineoffsets']
        else:
            # It appears matplotlib uses a different numbering if we only have 1 column, where x == 1 (normally the first column is x == 0). This makes me sad. Might be bug...
            return 1
    
    def annotate_ground_state(self):
        """
        Add a label to our ground state.
        
        This method is called automatically as part of the make() method.
        :return: Nothing.
        """
        # First work out where our GS is on the graph.
        x_pos = self.get_x_pos_from_multiplicity(self.ground_state.multiplicity)
        
        # Now plot our text.
        self.axes.annotate(
             r"$\bf{{{}}}_{{{}}}$".format(self.ground_state.multiplicity_symbol, self.ground_state.multiplicity_level),
             (x_pos, self.ground_state.energy),
             textcoords = "offset pixels",
             xytext = (0, 7),
             horizontalalignment = "center",
             fontsize = 12
         )
        
    def annotate_excited_state(self, state):
        """
        Add a label to an excited state
        
        This method is called automatically as part of the make() method.
        :param multiplicity_symbol: The symbol (a string) to annotate (eg, S1, T1).
        :return: Nothing.
        """
        # First work out where our state is on the graph.
        #state = self.excited_states.get_state(multiplicity_symbol)
        x_pos = self.get_x_pos_from_multiplicity(state.multiplicity)
        
        # Assemble our label.
        label_string =     r"$\bf{{{}}}_{{{}}}$: ".format(state.multiplicity_symbol, state.multiplicity_level) + "{:0.2f} eV".format(state.energy)
                        
        # Add oscillator strength (if we have it).
        if state.oscillator_strength is not None:
            label_string +=    "\n" + "f: {:0.2f}".format(state.oscillator_strength)
        
        # Now plot our text.
        self.axes.annotate(
             label_string,
             (x_pos, state.energy),
             textcoords = "offset pixels",
             xytext = (0, -7),
             verticalalignment = "top",
             horizontalalignment = "center",
             fontsize = 12
         )
    
    def annotate_dest(self, x, y):
        """
        """
        # First the text label.
        self.axes.annotate(
                r"$\bf{{ΔE_{{ST}}}}$: {:0.2f} eV".format(self.excited_states.singlet_triplet_energy),
                (x, y),
                textcoords = "offset pixels",
                xytext = (0,0),
                verticalalignment = "center",
                horizontalalignment = "left",
                fontsize = 12
            )
    
    def plot_labels(self):
        """
        Plot the labels on our graph that identify various excited states and the singlet/triplet splitting energy (dEst) if available.
        
        This method is called automatically as part of the make() method.
        :return: Nothing.
        """
        # First add a label for our ground state, unless there isn't room (this hack again).
        lowest_state = self.excited_states[0]
        # We need about 0.6 eV space.
        if (lowest_state.energy - 0.6) >= 0:
            self.annotate_ground_state()
        
        # Now we'll add a label for the lowest state of each multiplicity.
        for multiplicity in self.grouped_excited_states:
            self.annotate_excited_state(self.grouped_excited_states[multiplicity][0])
            
            
        # We'll also try and add a dEst label (obviously this is only possible if there is an S1 and T1 available).
        if self.show_dest:
            # First get our two states.
            S1 = self.excited_states.get_state("S(1)")
            T1 = self.excited_states.get_state("T(1)")
            
            # Find the midpoint between them.
            y_pos = min([S1.energy, T1.energy]) +  (abs(S1.energy - T1.energy) /2)
            
            # The label will be to the right of S1/T1.
            x_pos = max(self.get_x_pos_from_multiplicity(S1.multiplicity), self.get_x_pos_from_multiplicity(T1.multiplicity)) + (self.plot_options['linelengths'] /2) + 0.2
            # Add the label
            self.annotate_dest(x_pos, y_pos)
            
            # Now draw our vertical arrow.
            self.axes.annotate(
                "",
                xy = (x_pos-0.05, S1.energy),
                xytext = (x_pos-0.05, T1.energy),
                arrowprops=dict(arrowstyle="<->")
            )
            
            # Now extend a line from S1 and T1 to meet the vertical arrow.
            for state in (S1, T1):
                self.axes.annotate(
                    "",
                    xy = (self.get_x_pos_from_multiplicity(state.multiplicity) + (self.plot_options['linelengths'] /2), state.energy),
                    xytext = (x_pos-0.05, state.energy),
                    arrowprops=dict(arrowstyle="-", linestyle="--")
                )
            
            # Pad to the right so our dest label is chopped off.
            self.right_padding = 0.7            

        
    def x_axis_formatter(self, x, pos):
        """
        This method tells matplotlib how to label our X axis (with multiplicity labels).
        """
        # The multiplicities that we're using.
        mult = sorted(self.grouped_states)
        
        if len(self.grouped_states) > 1:
            # Check to see if our x is in our list of multiplicities.
            if x % self.plot_options['lineoffsets'] == 0 and x >= 0:
                try:
                    return digichem.result.excited_state.Excited_state.multiplicity_number_to_string((mult[int(x / self.plot_options['lineoffsets'])]))
                    
                except IndexError:
                    # Out of range.
                    return ""
        else:
            # Matplotlib uses different counting when there's only 1 column :(.
            if x == 1:
                return digichem.result.excited_state.Excited_state.multiplicity_number_to_string(mult[0])
            else:
                return ""
            
    def all_limits(self):
        """
        Limit the Y axis so all states are visible.
        """
        # Use auto scaling.
        self.axes.autoscale(enable = True, axis = 'y')
        
        # We'll go a small amount below 0 so our S0 is off the bottom of the graph.
        min_y = -0.1
        # Use a sneaky hack so that if the max energy is close to an integer, we'll use the next highest integer.
        #max_y = math.ceil(self.excited_states[-1].energy +0.1)
        max_y = self.excited_states[-1].energy + 0.1
        self.axes.set_ylim(min_y, max_y)
        
        # Adjusty hack.
        self._adjust_limits_for_zero()
        
    def auto_limits(self):
        """
        Limit the Y axis so S1, T1 ... N1 etc are all visible.
        """
        self.all_limits()
        
        # We'll show the lowest excited state of each mult.
        highest_state = max([self.grouped_excited_states[multiplicity][0].energy for multiplicity in self.grouped_excited_states])
        # Use a sneaky hack so that if the max energy is close to an integer, we'll use the next highest integer.
        max_y = math.ceil(highest_state +0.1)
        self.axes.set_ylim(None, max_y)
        
        
    def _adjust_limits_for_zero(self):
        """
        This method is a hack which lowers the y axis limit in cases where S1/T1 is very close to S0
        
        This is necessary because otherwise the state annotations will overflow the bottom of the diagram.
        """
        # First get our lowest E excited state.
        lowest_state = self.excited_states[0]
        # We need about 0.7 eV space.
        if (lowest_state.energy - 0.5) < 0:
            # Add on a bit extra.
            self.axes.set_ylim(bottom = lowest_state.energy - (0.5))
        
        
    def simple_limits(self, y_min = -0.1, y_max = 5):
        """
        Limit the Y axis between between two points.
        
        :param y_min: The start of the y axis.
        :param y_max: The end of the y axis.
        """
        self.axes.set_ylim(y_min, y_max)
        
    def adjust_axes(self):
        """
        Adjust the axes of our graph.
        
        This method is called automatically as part of the make() method.
        :return: Nothing.
        """
        super().adjust_axes()
                
        # Adjust axis
        # Add a bit o' padding (often we actually reduce padding from the default).
        xlim = self.axes.get_xlim()
        self.axes.set_xlim(xlim[0] - (self.left_padding -0.25), xlim[1] + (self.right_padding -0.25))
        #self.axes.set_xlim(xlim[0] +0.25, xlim[1] -0.25)
        
        # Now decide how big our y axis will be.
        if self.limits_method == "all":
            self.all_limits()
        elif self.limits_method == "auto":
            self.auto_limits()
        elif isinstance(self.limits_method, tuple) or isinstance(self.limits_method, list):
            self.simple_limits(self.limits_method[0], self.limits_method[1])
        else:
            raise ValueError("Unknown limits method '{}'".format(self.limits_method))
        
        # Set custom text on our x axis.
        self.axes.xaxis.set_major_formatter(FuncFormatter(self.x_axis_formatter))
        #self.axes.xaxis.set_major_locator(ticker.MultipleLocator(1))
        self.axes.xaxis.set_major_locator(ticker.FixedLocator([self.get_x_pos_from_multiplicity(multiplicity) for multiplicity in self.grouped_excited_states]))
        
        # Both our dimensions can scale to fit more data.
        self.figure.tight_layout()
        self.constant_scale(0, self.inch_per_x)
        self.constant_scale(1, self.inch_per_y)




# General imports.
from datetime import datetime, timedelta

from digichem.exception.base import Result_unavailable_error

# Digichem imports.
from digichem.parse.cclib import Cclib_parser
import digichem.log
import digichem.file.types as file_types

# Hidden imports.
#import pysoc.io.SOC

class Gaussian_parser(Cclib_parser):
    """
    Top level class for parsing output from Gaussian log files.
    """
    
    # A dictionary of recognised auxiliary file types.
    INPUT_FILE_TYPES = {
            file_types.gaussian_chk_file: "chk_file",
            file_types.gaussian_fchk_file: "fchk_file",
            file_types.gaussian_rwf_file: "rwf_file"
        }
    
    # Headers for date strings.
    DATE_HEADER = "Normal termination of"
    ELAPSED_TIME_HEADER = "Elapsed time:"
    CPU_TIME_HEADER = "Job cpu time:"
    CPU_HEADER = "Will use up to"

    def __init__(self, *log_files, rwfdump = "rwfdump", options, **auxiliary_files):
        self.rwfdump = rwfdump
        super().__init__(*log_files, options = options, **auxiliary_files)
    
    def parse_metadata(self):
        """
        Parse additional calculation metadata.
        """
        super().parse_metadata()
    
    def pre_parse(self):
        """
        Perform any setup before line-by-line parsing.
        """
        super().pre_parse()
        # Assume we used 1 CPU if not otherwise clear (is this a good idea?)
        self.data.metadata['num_cpus'] = 1
        
        self.wall_time = []
        self.cpu_time = []
    
    def parse_output_line(self, log_file, line):
        """
        Perform custom line-by-line parsing of an output file.
        """
        # Although we only need the last ~5 lines from the (possibly huge) log file, we read all the way through because negative seek()ing is tricky.
        # Look for our key string.
        if self.DATE_HEADER in line:
            # This line looks like: "Normal termination of Gaussian 16 at Sun Dec  6 19:13:09 2020"
            date_str = " ".join(line.split()[-4:])
            self.data.metadata['date'] = datetime.strptime(date_str, "%b %d %H:%M:%S %Y.").timestamp()
            
        elif self.ELAPSED_TIME_HEADER in line:
            # This line looks like: "Elapsed time:       0 days  2 hours 38 minutes 50.9 seconds."
            datey = line.split()[-8:]
            self.wall_time.append(timedelta(days = int(datey[0]), hours = int(datey[2]), minutes = int(datey[4]), seconds = float(datey[6])).total_seconds())
            
        elif self.CPU_TIME_HEADER in line:
            # This line looks like: "Job cpu time:       0 days 20 hours 52 minutes 17.3 seconds."
            datey = line.split()[-8:]
            self.cpu_time.append(timedelta(days = int(datey[0]), hours = int(datey[2]), minutes = int(datey[4]), seconds = float(datey[6])).total_seconds())
            
        elif self.CPU_HEADER in line:
            # This line looks like: "Will use up to   10 processors via shared memory."
            self.data.metadata['num_cpus'] = int(line.split()[4])
            
    def post_parse(self):
        """
        Perform any required operations after line-by-line parsing.
        """
        super().post_parse()
        
        if 'wall_time' not in self.data.metadata and len(self.wall_time) != 0:
            self.data.metadata['wall_time'] = self.wall_time
        
        if 'cpu_time' not in self.data.metadata and len(self.cpu_time) != 0:
            self.data.metadata['cpu_time'] = self.cpu_time
        
        # Get SOC.
        # Next try and get SOC.
        try:
            self.calculate_SOC()
            
        except Exception:
            digichem.log.get_logger().debug("Cannot calculate spin-orbit-coupling from output file '{}'".format(self.log_file_path), exc_info = True)
    
    def calculate_SOC(self):
        """
        Parse spin-orbit coupling using PySOC.
        """
        try:
            import pysoc.io.SOC
        
        except Exception as e:
            raise Result_unavailable_error("Spin-orbit coupling", "PySOC is not available") from e
        
        # For SOC, we need both .log and .rwf file.
        # No need to check for these tho; pysoc does that for us.
        # We also need etsyms to decide which excited state is which.
        if not hasattr(self.data, "etsyms"):
            raise Result_unavailable_error("Spin-orbit coupling", "There are no excited states available")
        
        # Get a PySOC parser.
        soc_calculator = pysoc.io.SOC.Calculator(self.log_file_path, rwfdump = self.rwfdump, rwf_file_name = self.auxiliary_files['rwf_file'])
        soc_calculator.calculate()
        SOC_table = soc_calculator.soc_td.SOC
        
        # We'll split the SOC table given to use by PySOC to better match the format used by cclib.
        socstates = []
        socelements = []
        
        for SOC_line in SOC_table:
            # Add states.
            socstates.append([SOC_line.singlet_state, SOC_line.triplet_state])
            
            # Add coupling.
            socelements.append([SOC_line.positive_one, SOC_line.zero, SOC_line.negative_one])
                
        # Add to data.
        self.data.socstates = socstates
        self.data.socelements = socelements
            
            


# Exceptions and other errors.

class Digichem_exception(Exception):
    """
    General digichem exception.
    """
    
    def __init__(self, message):
        self.message = message
        
    def __str__(self, *args, **kwargs):
        return self.message
    

class Result_unavailable_error(Digichem_exception):
    """
    Exception for when a requested result is not available (because it could not be found in the calculation results for example).
    """
    
    def __init__(self, result_name, reason = "result could not be found"):
        """
        Constructor for Result_unavailable_error objects.
        
        :param result_name: The name of the result that is unavailable.
        :param reason: Optional message explaining why the result is unavailable. If not given, a default message will be used.
        """
        self.result_name = result_name
        self.reason = reason
        
    def __str__(self, *args, **kwargs):
        """
        Stringify this error.
        """
        return "'{}' is not available; {}".format(self.result_name, self.reason)
        
class File_maker_exception(Digichem_exception):
    """
    Exception for when a file cannot be made/rendered for whatever reason.
    """
    
    def __init__(self, file_maker, reason = ""):
        """
        Constructor for File_maker_exception objects.
        
        :param file_maker: The file_maker object where the exception occurred.
        :param reason: Optional string describing why the exception occurred.
        """
        self.file_maker = file_maker
        self.reason = reason
        
    def __str__(self, *args, **kwargs):
        """
        Stringify this error.
        """
        return "Error making '{}' file '{}'; {}".format(type(self.file_maker).__name__, self.file_maker.output, self.reason)
    
class Unknown_file_type_exception(Digichem_exception):
    """
    Exception for when a file is given but its type cannot be determined.
    """
    
    def __init__(self, file_path, expected = None):
        """
        Constructor for Unknown_file_type_exception objects.
        
        :param file_path: String-like path of the file that is unrecognised.
        :param expected: An optional string-like representing the type of file that was expected. 
        """
        self.file_path = file_path
        self.expected = expected
        
    def __str__(self, *args, **kwargs):
        """
        Stringify this error.
        """
        err_str = "Unknown file type '{}'".format(self.file_path)
        if self.expected is not None:
            err_str = "{}; expected file of type '{}'".format(err_str, self.expected)
        return err_str
                
class Submission_error(Digichem_exception):
    """
    Exceptions for when an error occurs during calculation submission.
    """
    
    def __init__(self, calculation, reason, file_name = None):
        """
        Constructor for Submission_error exception objects.
        
        :param calculation: The calculation that was in process of being submitted when the error occurred. This can be any one of a Method_target, Program_target or Calculation_target.
        :param reason: String describing why the error occurred.
        """        
        # Do some quick type checking.
        if calculation.meta['TYPE'] == "destination":
            # 'Calculation' is actually a Method_target.
            calculation = calculation.program.calculation
        elif calculation.meta['TYPE'] == "program":
            # 'Calculation' is actually a Program_target.
            calculation = calculation.calculation
        
        # Decide on file name.
        if file_name is None:
            file_name = calculation.molecule_name
        self.file_name = file_name
        
        self.calculation = calculation
        self.reason = reason
        
    def __str__(self, *args, **kwargs):
        """
        Stringify this error.
        """
        return "Error submitting file '{}' to '{}'; {}".format(self.file_name, self.calculation.meta['name'], self.reason)
    
    
class Format_error(Digichem_exception):
    """
    Exceptions for when an error occurs during result formatting.
    """
    
    def __init__(self, format, reason):
        """
        Constructor for Format_error exception objects.
        """
        self.format = format
        self.reason = reason
        
    def __str__(self):
        """
        Stringify this error.
        """
        return "{} ({}); {}".format(type(self.format).__name__, ", ".join(getattr(self.format, 'CLASS_HANDLE', [])), self.reason)
    


import pytest
from pathlib import Path
import shutil

from digichem.image.excited_states import Excited_states_diagram_maker
from digichem.image.graph import Convergence_graph_maker
from digichem.image.orbitals import Orbital_diagram_maker
from digichem.image.spectroscopy import Absorption_graph_maker, Emission_graph_maker, \
    Frequency_graph_maker, NMR_graph_maker, NMR_graph_zoom_maker
from digichem.image.structure import Skeletal_image_maker
import digichem.image.vmd
import digichem.image.render


from digichem.parse.util import open_for_parsing, parse_calculation

from digichem.test.util import digichem_options, data_directory
from digichem.test.test_result import gaussian_ES_result, turbomole_ES_result, orca_ES_result, \
    gaussian_opt_result, turbomole_opt_result, orca_opt_result, orca_opt_freq_result, \
    orca_nmr_result

HAS_VMD = shutil.which(digichem.config.get_config()['render']['vmd']['executable'])
HAS_BLENDER = shutil.which(digichem.config.get_config()['render']['batoms']['blender'])


@pytest.mark.parametrize("result_set", [
    pytest.lazy_fixture("gaussian_ES_result"),
    pytest.lazy_fixture("turbomole_ES_result"),
    pytest.lazy_fixture("orca_ES_result")
])
def test_es_diagram(result_set, tmp_path, digichem_options):
    """Can we make an excited states diagram?"""
    maker = Excited_states_diagram_maker.from_options(
        tmp_path / "tmp.png",
        excited_states = result_set.excited_states,
        ground_state = result_set.ground_state,
        options = digichem_options
    )

    maker.get_image()
    assert Path(tmp_path, "tmp.png").exists()


@pytest.mark.parametrize("result_set", [
    pytest.lazy_fixture("gaussian_opt_result"),
    pytest.lazy_fixture("turbomole_opt_result"),
    pytest.lazy_fixture("orca_opt_result")
])
def test_convergence_diagram(result_set, tmp_path, digichem_options):
    """Can we make an optimisation graph?"""
    maker = Convergence_graph_maker.from_options(
        tmp_path / "tmp.png",
        energies = result_set.energies.scf,
        options = digichem_options
    )

    maker.get_image()
    assert Path(tmp_path, "tmp.png").exists()


@pytest.mark.parametrize("result_set", [
    pytest.lazy_fixture("gaussian_opt_result"),
    pytest.lazy_fixture("turbomole_opt_result"),
    pytest.lazy_fixture("orca_opt_result")
])
def test_orbital_diagram(result_set, tmp_path, digichem_options):
    """Can we make an orbital diagram?"""
    maker = Orbital_diagram_maker.from_options(
        tmp_path / "tmp.png",
        orbitals = result_set.orbitals,
        options = digichem_options
    )

    maker.get_image()
    assert Path(tmp_path, "tmp.png").exists()


@pytest.mark.parametrize("result_set", [
    pytest.lazy_fixture("gaussian_ES_result"),
    pytest.lazy_fixture("turbomole_ES_result"),
    pytest.lazy_fixture("orca_ES_result")
])
def test_abs_diagram(result_set, tmp_path, digichem_options):
    """Can we make an absorption spectrum?"""
    maker = Absorption_graph_maker.from_options(
        tmp_path / "tmp.png",
        excited_states = result_set.excited_states,
        options = digichem_options
    )

    maker.get_image()
    assert Path(tmp_path, "tmp.png").exists()


@pytest.mark.parametrize("result_set", [
    pytest.lazy_fixture("gaussian_ES_result"),
    pytest.lazy_fixture("turbomole_ES_result"),
    pytest.lazy_fixture("orca_ES_result")
])
def test_pl_diagram(result_set, tmp_path, digichem_options):
    """Can we make an emission spectrum?"""
    maker = Emission_graph_maker.from_options(
        tmp_path / "tmp.png",
        excited_states = result_set.excited_states,
        options = digichem_options
    )

    maker.get_image()
    assert Path(tmp_path, "tmp.png").exists()


@pytest.mark.parametrize("result_set", [
    pytest.lazy_fixture("gaussian_opt_result"),
    pytest.lazy_fixture("turbomole_opt_result"),
    pytest.lazy_fixture("orca_opt_freq_result")
])
def test_freq_diagram(result_set, tmp_path, digichem_options):
    """Can we make an IR spectrum?"""
    maker = Frequency_graph_maker.from_options(
        tmp_path / "tmp.png",
        vibrations = result_set.vibrations,
        options = digichem_options
    )

    maker.get_image()
    assert Path(tmp_path, "tmp.png").exists()


@pytest.mark.parametrize("result_set", [
    pytest.lazy_fixture("orca_nmr_result"),
])
def test_nmr_diagram(result_set, tmp_path, digichem_options):
    """Can we make an NMR spectrum?"""
    # Get a suitable spectrometer object.
    spectrometer = result_set.nmr.spectrometer

    maker = NMR_graph_maker.from_options(
        tmp_path / "tmp.png",
        graph = spectrometer.get_graph(1, 1),
        options = digichem_options
    )

    maker.get_image()
    assert Path(tmp_path, "tmp.png").exists()


@pytest.mark.parametrize("result_set", [
    pytest.lazy_fixture("orca_nmr_result"),
])
def test_focus_nmr_diagram(result_set, tmp_path, digichem_options):
    """Can we make an NMR spectrum?"""
    # Get a suitable spectrometer object.
    spectrometer = result_set.nmr.spectrometer

    maker = NMR_graph_zoom_maker.from_options(
        tmp_path / "tmp.png",
        graph = spectrometer.get_graph(1, 1),
        focus = result_set.atoms.groups[(2, "H")],
        options = digichem_options
    )

    maker.get_image()
    assert Path(tmp_path, "tmp.png").exists()


@pytest.mark.parametrize("result_set", [
    pytest.lazy_fixture("gaussian_opt_result"),
    pytest.lazy_fixture("turbomole_opt_result"),
    pytest.lazy_fixture("orca_opt_freq_result")
])
def test_2d_diagram(result_set, tmp_path, digichem_options):
    """Can we make an IR spectrum?"""
    maker = Skeletal_image_maker.from_options(
        tmp_path / "tmp.png",
        atoms = result_set.atoms,
        options = digichem_options
    )

    maker.get_image()
    assert Path(tmp_path, "tmp.png").exists()


@pytest.mark.parametrize("cube_file", [
    Path(data_directory(), "Cubes/Pyridine.HOMO.cube")
], ids = ["Gaussian"])
@pytest.mark.parametrize("maker_cls", [
    pytest.param(digichem.image.vmd.Structure_image_maker, marks = pytest.mark.skipif(not HAS_VMD, reason = "VMD is not available")),
    pytest.param(digichem.image.vmd.Orbital_image_maker, marks = pytest.mark.skipif(not HAS_VMD, reason = "VMD is not available")),
    pytest.param(digichem.image.render.Structure_image_maker, marks = pytest.mark.skipif(not HAS_BLENDER, reason = "Blender is not available")),
    pytest.param(digichem.image.render.Orbital_image_maker, marks = pytest.mark.skipif(not HAS_BLENDER, reason = "Blender is not available")),
])
def test_3d_image(cube_file, maker_cls, tmp_path, digichem_options):
    """Can we make a 3D structure image?"""
    maker = maker_cls.from_options(
        tmp_path / "tmp.png",
        cube_file = cube_file,
        options = digichem_options
    )

    maker.get_file('x0y0z0')
    assert Path(tmp_path, "tmp.x0y0z0.png").exists()


@pytest.mark.parametrize("cube_file", [
    Path(data_directory(), "Cubes/Pyridine.SCF.cube")
], ids = ["Gaussian"])
@pytest.mark.parametrize("maker_cls", [
    pytest.param(digichem.image.vmd.Density_image_maker, marks = pytest.mark.skipif(not HAS_VMD, reason = "VMD is not available")),
    pytest.param(digichem.image.render.Density_image_maker, marks = pytest.mark.skipif(not HAS_BLENDER, reason = "Blender is not available")),
], ids = ["vmd", "batoms"])
def test_density_image(cube_file, maker_cls, tmp_path, digichem_options):
    """Can we make a 3D structure image?"""

    maker = maker_cls.from_options(
        tmp_path / "tmp.png",
        cube_file = cube_file,
        options = digichem_options
    )

    maker.get_file('x0y0z0')
    assert Path(tmp_path, "tmp.x0y0z0.png").exists()


@pytest.mark.parametrize("cube_file", [
    Path(data_directory(), "Cubes/Benzene.anion.spin.cube")
], ids = ["Gaussian"])
@pytest.mark.parametrize("maker_cls", [
    pytest.param(digichem.image.vmd.Spin_density_image_maker, marks = pytest.mark.skipif(not HAS_VMD, reason = "VMD is not available")),
    pytest.param(digichem.image.render.Spin_density_image_maker, marks = pytest.mark.skipif(not HAS_BLENDER, reason = "Blender is not available")),
], ids = ["vmd", "batoms"])
def test_spin_density_image(cube_file, maker_cls, tmp_path, digichem_options):
    """Can we make a 3D structure image?"""
    maker = maker_cls.from_options(
        tmp_path / "tmp.png",
        cube_file = cube_file,
        options = digichem_options
    )

    maker.get_file('x0y0z0')
    assert Path(tmp_path, "tmp.x0y0z0.png").exists()


@pytest.mark.parametrize("homo_cube, lumo_cube", [
    [Path(data_directory(), "Cubes/Pyridine.HOMO.cube"), Path(data_directory(), "Cubes/Pyridine.LUMO.cube")]
], ids = ["Gaussian"])
@pytest.mark.parametrize("maker_cls", [
    pytest.param(digichem.image.vmd.Combined_orbital_image_maker, marks = pytest.mark.skipif(not HAS_VMD, reason = "VMD is not available")),
    pytest.param(digichem.image.render.Combined_orbital_image_maker, marks = pytest.mark.skipif(not HAS_BLENDER, reason = "Blender is not available")),
], ids = ["vmd", "batoms"])
def test_combined_orbital_image(homo_cube, maker_cls, lumo_cube, tmp_path, digichem_options):
    """Can we make a 3D structure image?"""
    maker = maker_cls.from_options(
        tmp_path / "tmp.png",
        HOMO_cube_file = homo_cube,
        LUMO_cube_file = lumo_cube,
        options = digichem_options
    )

    maker.get_file('x0y0z0')
    assert Path(tmp_path, "tmp.x0y0z0.png").exists()


@pytest.mark.parametrize("cube_file", [
    Path(data_directory(), "Cubes/Benzene.anion.AHOMO.cube")
], ids = ["Gaussian"])
@pytest.mark.parametrize("maker_cls", [
    pytest.param(digichem.image.vmd.Alpha_orbital_image_maker, marks = pytest.mark.skipif(not HAS_VMD, reason = "VMD is not available")),
    pytest.param(digichem.image.render.Alpha_orbital_image_maker, marks = pytest.mark.skipif(not HAS_BLENDER, reason = "Blender is not available")),
], ids = ["vmd", "batoms"])
def test_unrestricted_orbital_image(cube_file, maker_cls, tmp_path, digichem_options):
    """Can we make a 3D structure image?"""
    maker = maker_cls.from_options(
        tmp_path / "tmp.png",
        cube_file = cube_file,
        options = digichem_options
    )

    maker.get_file('x0y0z0')
    assert Path(tmp_path, "tmp.x0y0z0.png").exists()


@pytest.mark.parametrize("result_path, cube_file", [
    [  
        Path(data_directory(), "Pyridine/Gaussian 16 Optimisation Frequencies PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz"),
        Path(data_directory(), "Cubes/Pyridine.HOMO.cube"),
    ]
])
@pytest.mark.parametrize("maker_cls", [
    pytest.param(digichem.image.vmd.Permanent_dipole_image_maker, marks = pytest.mark.skipif(not HAS_VMD, reason = "VMD is not available")),
    pytest.param(digichem.image.render.Permanent_dipole_image_maker, marks = pytest.mark.skipif(not HAS_BLENDER, reason = "Blender is not available")),
], ids = ["vmd", "batoms"])
def test_pdm_image(result_path, cube_file, maker_cls, tmp_path, digichem_options):
    """Can we make a 3D structure image?"""
    with open_for_parsing(result_path) as open_files:
        result_set = parse_calculation(*open_files, options = digichem_options)
        maker = maker_cls.from_options(
            tmp_path / "tmp.png",
            cube_file = cube_file,
            dipole_moment = result_set.dipole_moment,
            options = digichem_options
        )

        maker.get_file('x0y0z0')
        assert Path(tmp_path, "tmp.x0y0z0.png").exists()


@pytest.mark.parametrize("result_path, cube_file", [
    [
        Path(data_directory(), "Pyridine/Gaussian 16 Excited States TDA Optimised S(1) PBE1PBE (GD3BJ) Toluene 6-31G(d,p).tar.gz"),
        Path(data_directory(), "Cubes/Pyridine.HOMO.cube"),
    ]
])
@pytest.mark.parametrize("maker_cls", [
    pytest.param(digichem.image.vmd.Transition_dipole_image_maker, marks = pytest.mark.skipif(not HAS_VMD, reason = "VMD is not available")),
    pytest.param(digichem.image.render.Transition_dipole_image_maker, marks = pytest.mark.skipif(not HAS_BLENDER, reason = "Blender is not available")),
], ids = ["vmd", "batoms"])
def test_tdm_image(result_path, cube_file, maker_cls, tmp_path, digichem_options):
    """Can we make a 3D structure image?"""
    with open_for_parsing(result_path) as open_files:
        result_set = parse_calculation(*open_files, options = digichem_options)
        maker = maker_cls.from_options(
            tmp_path / "tmp.png",
            cube_file = cube_file,
            dipole_moment = result_set.excited_states.find("S(1)").transition_dipole_moment.electric,
            magnetic_dipole_moment = result_set.excited_states.find("S(1)").transition_dipole_moment.magnetic,
            options = digichem_options
        )

        maker.get_file('x0y0z0')
        assert Path(tmp_path, "tmp.x0y0z0.png").exists()


@pytest.mark.parametrize("cube_file", [
    Path(data_directory(), "Cubes/Pyridine.HOMO.cube"),
])
@pytest.mark.parametrize("maker_cls", [
    pytest.param(digichem.image.vmd.Orbital_image_maker, marks = pytest.mark.skipif(not HAS_VMD, reason = "VMD is not available")),
    pytest.param(digichem.image.render.Orbital_image_maker, marks = pytest.mark.skipif(not HAS_BLENDER, reason = "Blender is not available")),
], ids = ["vmd", "batoms"])
def test_nto_image(cube_file, maker_cls, tmp_path, digichem_options):
    """Can we make a 3D structure image?"""
    maker = maker_cls.from_options(
        tmp_path / "tmp.png",
        cube_file = cube_file,
        options = digichem_options
    )

    maker.get_file('x0y0z0')
    assert Path(tmp_path, "tmp.x0y0z0.png").exists()


# Hidden import for speed.
#import colour
# from digichem.result.spectroscopy import Spectroscopy_graph,\
#     Absorption_emission_graph

# General imports.
from itertools import filterfalse, zip_longest
import numpy
import math
import warnings
import itertools

from digichem.misc.text import andjoin
from digichem.exception.base import Result_unavailable_error
from digichem.result import Result_container
from digichem.result import Result_object
from digichem.result.base import Floatable_mixin
import digichem.log


class Excited_state_list(Result_container):
    """
    Class for representing a group of excited states.
    """
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
    @property
    def singlet_triplet_energy(self):
        """
        Get ΔE st; the energy difference between the T1 and S1 excited states.
        
        The returned value is positive if S1 is higher in energy than T1 (the usual case), negative otherwise.
        
        :return: The singlet-triplet splitting energy (in eV).
        """
        # Get our two states of interest.
        S1 = self.get_state("S(1)")
        T1 = self.get_state("T(1)")
        # Calculate their difference.))
        return float(S1) - float(T1)
    
    @property
    def multiplicity_strings(self):
        """
        Get a string describing the different multiplicities of the excited states of this list.
        """
        group = self.group()
        mults = []
        for excited_states in group.values():
            mults.append(excited_states[0].multiplicity_string)
            
        return andjoin(mults)
    
    def difference(self, state1, state2):
        """
        Get the difference in energy between the states by given labels.
        
        The returned energy difference will always be positive.
        
        :param state1: State symbol of the first state.
        :param state2: State symbol of the second state.
        """
        states = (self.get_state(state_symbol = state1).energy, self.get_state(state_symbol = state2).energy)
        return max(states) - min(states)
    
    def find(self, criteria = None, *, state_symbol = None, level = None, mult_level = None):
        """
        Retrieve a particular excited state from its state symbol (S(1), T(1) etc.).
        
        For multiplicities from 1 -> 4, single capital letters are used ('S', 'D', 'T' or 'Q' respectively). For higher multiplicities, the integer representation is used ('5', '6', '7' and so on).
        The multiplicity level follows in brackets, eg ('S(1)') refers to the first (lowest energy) singlet state.
        
        See Excited_state.state_symbol() for a full description of possible symbols.
        
        :raises ValueError: If the requested symbol could not be found in this list.
        :param criteria: Automatically determine which criteria to search by.
        :param state_symbol: The symbol (a string) to retrieve.
        :param level: The level (an int or string that looks like an int) to retrieve.
        :param mult_level: The multiplicity and multiplicity level (as a tuple, eg '(1,1)' for S(1), '(3,2)' for T(2)) to retrieve.
        :return: The requested excited state object.
        """
        if criteria is not None:
            if isinstance(criteria, tuple):
                mult_level = criteria
            elif criteria.isdigit() or isinstance(criteria, int):
                level = int(criteria)
            else:
                state_symbol = criteria
        
        # Now get our search func.
        if state_symbol is not None:
            filter_func = lambda state: state.state_symbol != state_symbol
        elif level is not None:
            filter_func = lambda state: state.level != level
        elif mult_level is not None:
            filter_func = lambda state: state.multiplicity != mult_level[0] or state.multiplicity_level != mult_level[1]
        else:
            raise ValueError("Missing criteria to search by; specify one of 'criteria', 'state_symbol', 'level' or 'mult_level'")
        
        try:
            return next(filterfalse(filter_func, self))
        except Exception:
            # Determine what we searched by for better error reporting.
            if state_symbol is not None:
                criteria_string = "symbol = '{}'".format(state_symbol)
                
            elif level is not None:
                criteria_string = "level = '{}'".format(level)
                
            else:
                criteria_string = "mult_level = '{}'".format(mult_level)
                
            raise Result_unavailable_error("Excited state", "could not find excited state with {}".format(criteria_string)) from None
        
    def get_state(self, *args, **kwargs):
        warnings.warn("get_state() is deprecated use find() instead", DeprecationWarning)
        return self.find(*args, **kwargs)
    
    def group(self):
        """
        Group the excited states in this list by multiplicity.
        
        :return: A dictionary of grouped excited states. Each key will correspond to a multiplicity (1, 2, 3 etc) and each value will be an Excited_state_list of states with that multiplicity. 
        """
        # Dictionary of grouped states.
        grouped_excited_states = {}
        
        for excited_state in self:
            # Group each ES by multiplicity.
            try:
                # Try and append to our list.
                grouped_excited_states[excited_state.multiplicity].append(excited_state)
            except KeyError:
                # No list exists yet.
                grouped_excited_states[excited_state.multiplicity] = type(self)([excited_state])
        
        # Sorting hack for old version of python that didn't technically support sorted dicts (although the sort order was secretly maintained).
        # Just make a new dict that is sorted from the start:
        sorted_group = {}
        for mult in sorted(grouped_excited_states):
            sorted_group[mult] = grouped_excited_states[mult]
        
        return sorted_group
    
    def group_pairs(self):
        """
        Return all the unique pair combinations of the different multiplicities of this list.
        
        :returns: A two membered tuple, where the first element is a list of pairs of multiplicities [(1, 2), (1, 3), (2, 3)] etc, and the second is a grouped dictionary of multiplicites (see group()).
        """
        group = self.group()
        return (list(itertools.combinations(group, 2)), group)
    
    @property
    def num_singlets(self):
        """
        The number of singlet excited states in this list.
        """
        grouped_states = self.group()
        return len(grouped_states[1]) if 1 in grouped_states else 0
    
    @property
    def num_triplets(self):
        """
        The number of triplet excited states in this list.
        """
        grouped_states = self.group()
        return len(grouped_states[3]) if 3 in grouped_states else 0
        
    @classmethod
    def from_parser(self, parser):
        """
        Create an Excited_state_list object from an output file parser.
        
        :param parser: An output file parser.
        :return: The populated Excited_state_list object.
        """
        return self(Excited_state.list_from_parser(parser))
    
    def assign_levels(self):
        """
        (Re)assign total and multiplicity levels of the excited states in this list.
        
        The contents of this list will be modified in place.
        """
        # A dictionary of multiplicities that we've seen so far.
        multiplicities = {}
        total_level = 0
        
        for state in self:
            # (try) and add this state's multiplicity to our dict (if we get a key error that just means it's the first time we've seen this mult).
            try:
                multiplicities[state.multiplicity] += 1
            except KeyError:
                # First time we've seen this mult, set to 1.
                multiplicities[state.multiplicity] = 1
            
            # Increment total level.
            total_level += 1
                
            # Set mult and total level.
            state.multiplicity_level = multiplicities[state.multiplicity]
            state.level = total_level

    @classmethod
    def merge(self, *multiple_lists, **kwargs):
        """
        Merge multiple lists of of the same type into a single, ordered list.
         
        Note that this method will return a new list object, but will modify the contained objects (eg, object.level) in place.
        Inheriting classes may safely override this method.
        """
        # A dictionary where each key is a multiplcity and each value is a list of excited states with that mult.
        multiplicities = {}
        for excited_states in multiple_lists:
            group = excited_states.group()
            for multiplicity in group:
                if multiplicity not in multiplicities:
                    # Firt time we've seen this mult, add.
                    multiplicities[multiplicity] = group[multiplicity]
                
                else:
                    # Already got this type of excited states.
                    warnings.warn("Attempting to merge excited states of multiplicity '{}' but this multiplicity has already been provided by an earlier result, ignoring".format(multiplicity))
        
        merged = self(itertools.chain(*multiplicities.values()), **kwargs)
        merged.sort()
        merged.assign_levels()
        return merged
    
    def _get_dump_(self):
        """
        Method used to get a dictionary used to generate on-demand values for dumping.
        
        This functionality is useful for hiding expense properties from the normal dump process, while still exposing them when specifically requested.
        
        Each key in the returned dict is the name of a dumpable item, each value is a function to call with digichem_options as its only param.
        """
        return {"spectrum": self.generate_spectrum}
    
    def generate_spectrum(self, digichem_options):
        """
        Abstract function that is called to generate an on-demand value for dumping.
        
        This functionality is useful for hiding expensive properties from the normal dump process, while still exposing them when specifically requested.
        
        :param key: The key being requested.
        :param digichem_options: Digichem options being used to dump.
        """
        from digichem.result.spectroscopy import Spectroscopy_graph, Absorption_emission_graph
        
        
        # Get spectrum data.
        # For excited states, we dump two spectra. One in nm, one in eV (which have different scaling).
        # TODO: It's weird that these spectra are only available in dumped format, there should be some property/function on the class that also returns them...
        spectrum_nm = Absorption_emission_graph.from_excited_states(
            self,
            fwhm = digichem_options['absorption_spectrum']['fwhm'],
            resolution = digichem_options['absorption_spectrum']['gaussian_resolution'],
            cutoff = digichem_options['absorption_spectrum']['gaussian_cutoff'],
            filter = digichem_options['absorption_spectrum']['y_filter'],
            use_jacobian = digichem_options['absorption_spectrum']['use_jacobian']
        )
        
        
        spectrum_ev = Spectroscopy_graph(
            [(excited_state.energy, excited_state.oscillator_strength) for excited_state in self],
            fwhm = digichem_options['absorption_spectrum']['fwhm'],
            resolution = digichem_options['absorption_spectrum']['gaussian_resolution'],
            cutoff = digichem_options['absorption_spectrum']['gaussian_cutoff'],
            filter = digichem_options['absorption_spectrum']['y_filter']
        )
        
        try:
            spectrum_nm_data = spectrum_nm.plot_cumulative_gaussian()
            y_units = "arb. unit" if digichem_options['absorption_spectrum']['use_jacobian'] else "oscillator_strength"
            
            spectrum_nm_data = [{"x":{"value": float(x), "units": "nm"}, "y": {"value":float(y), "units": y_units}} for x,y in spectrum_nm_data]
            spectrum_nm_peaks = [{"x":{"value": float(x), "units": "nm"}, "y": {"value":float(y), "units": y_units}} for x, y in spectrum_nm.peaks()]
        
        except Exception:
            spectrum_nm_data = []
            spectrum_nm_peaks = []
        
        try:
            spectrum_ev_data = spectrum_ev.plot_cumulative_gaussian()
            
            spectrum_ev_data = [{"x":{"value": float(x), "units": "eV"}, "y": {"value":float(y), "units": "oscillator_strength"}} for x,y in spectrum_ev_data]
            spectrum_ev_peaks = [{"x":{"value": float(x), "units": "eV"}, "y": {"value":float(y), "units": "oscillator_strength"}} for x, y in spectrum_ev.peaks()]
        
        except Exception:
            spectrum_ev_data = []
            spectrum_ev_peaks = []
            
        return {
            "nm": {
                "values": spectrum_nm_data,
                "peaks": spectrum_nm_peaks
            },
            "ev": {
                "values": spectrum_ev_data,
                "peaks": spectrum_ev_peaks
            }    
        }
    
    def _dump_(self, digichem_options, all):
        dump_dict = {
            "values": super()._dump_(digichem_options, all),
        }
        
        # Add extra properties.
        mult_pairs, mults = self.group_pairs()
        
        # Number of states of each multiplicity.
        for mult, states in mults.items():
            dump_dict['num_{}'.format(Energy_state.multiplicity_number_to_string(mult))] = len(states)
            
        # DeST and other values.
        for mult_pair in mult_pairs:
            state1 = mults[mult_pair[0]][0]
            state2 = mults[mult_pair[1]][0]
            dump_dict['dE({}{})'.format(state1.multiplicity_symbol, state2.multiplicity_symbol)] = {
                "value": float(state1.energy - state2.energy),
                "units": "eV"
            }
            
        return dump_dict
    
    @classmethod
    def from_dump(self, data, result_set, options):
        """
        Get a list of instances of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        return self(Excited_state.list_from_dump(data['values'], result_set, options))


class Excited_state_transition(Result_object):
    """
    Class that represents a transition that contributes to a particular excited state.
    """
    
    def __init__(self, level, starting_mo, ending_mo, coefficient):
        """
        Constructor for excited state transitions.
        
        :param level: The 'level' of this transition. The most significant (highest probability) transition has level 1.
        :param starting_mo: The Molecular_orbital object from which this transition begins.
        :param ending_mo: The Molecular_orbital object to which the transition ends.
        :param coefficient: The coefficient of this orbital. Square to get the probability.
        """
        self.level = level
        self.starting_mo = starting_mo
        self.ending_mo = ending_mo
        self.coefficient = coefficient
        
    @property
    def probability(self):
        """
        The probability of this transition.
        """
        return self.coefficient **2
    
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        return {
            "index": self.level,
            "start": {
                "spin": self.starting_mo.spin_type,
                "index": self.starting_mo.level,
                "label": self.starting_mo.label
            },
            "end": {
                "spin": self.ending_mo.spin_type,
                "index": self.ending_mo.level,
                "label": self.ending_mo.label
            },
            "coefficient": float(self.coefficient),
            "probability": float(self.probability **2)
        }
        
    @classmethod
    def list_from_parser(self, parser):
        """
        Create a list of excited state transitions from an output file parser.
        
        :param parser: An output file parser.
        :param alpha_mo_list: A Molecular_orbital_list object of the MOs of this system.
        :param beta_mo_list: A Molecular_orbital_list object of the beta MOs of this system. This can be left as null for restricted calcs.
        :return: A list of Excited_state_transition objects.
        """
        try:
            # Create a tuple of our MOs (helps us later).
            MOs = (parser.results.orbitals, parser.results.beta_orbitals)
            
            transitions_list = []
            
            # etsecs is a list of list, the outer list corresponds to each excited state, the inner list contains actual transitions.
            for excited_state_transitions in parser.data.etsecs:
    
                # We'll first create an intermediate list of keyword dicts which we'll then sort.
                data_list = []

                for (starting_mo_index, starting_mo_AB), (ending_mo_index, ending_mo_AB), coefficient in excited_state_transitions:
                    try:
                        data_list.append({
                            'starting_mo': MOs[starting_mo_AB][starting_mo_index],
                            'ending_mo': MOs[ending_mo_AB][ending_mo_index],
                            'coefficient': coefficient
                        })
                    
                    except IndexError:
                        # This is fairly common in Orca 6, where only a subset of virtual orbitals are printed by default.
                        digichem.log.get_logger() \
                            .warning("Unable to construct excited state transition; transition is to/from an orbital that is not available ({} and {})".format(starting_mo_index, ending_mo_index))
                
                # Sort by probability/coefficient.
                data_list.sort(key=lambda keywords: math.fabs(keywords['coefficient']), reverse=True)
                
                # Now get a list objects and append to our big list.
                transitions_list.append(
                    [self(index+1, keywords['starting_mo'], keywords['ending_mo'], keywords['coefficient']) for index, keywords in enumerate(data_list)]
                )
                
            # All done.
            return transitions_list
        
        except AttributeError:
            # No data.
            return []
        
    @classmethod
    def list_from_dump(self, data, result_set, options):
        """
        Get a list of instances of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        trans = []
        for tran_dict in data:
            # Get our start and end MOs.
            orbitals = {}
            for key in ("start", "end"):
                if tran_dict[key]['spin'] in ("none", "alpha"):
                    orbitals[key] = result_set.orbitals[tran_dict[key]['index'] -1]
                
                else:
                    orbitals[key] = result_set.beta_orbitals[tran_dict[key]['index'] -1]
            
            trans.append(self(tran_dict['index'], orbitals['start'], orbitals['end'], tran_dict['coefficient']))
        
        return trans


class Energy_state(Result_object, Floatable_mixin):
    """
    Class for representing different energy states of the same system.
    """
    
    # Colour categories.
    colors = [
        {"max": 400, "name": "Ultraviolet"},
        {"max": 420, "name": "Violet"},
        {"max": 470, "name": "Blue"},
        {"max": 505, "name": "Cyan"},
        {"max": 555, "name": "Green"},
        {"max": 595, "name": "Yellow"},
        {"max": 625, "name": "Orange"},
        {"max": 740, "name": "Red"},
        {"max": float("inf"), "name": "Infrared"}    
    ]
    
    def __init__(self, level, multiplicity, multiplicity_level, energy):
        """
        Constructor for Energy_state objects.
        
        :param level: The ordered (by energy) index of this energy state, where 0 is the GS and 1 is the lowest excited state.
        :param multiplicity_level: the ordered (by energy) index of this energy state in terms of states that share the same multiplicity.
        :param multiplicity: The multiplicity of this state as a number (eg, 1 for singlet, 2 for doublet). Fractional multiplicities are also accepted.
        :param energy: The energy of this state in eV. Whether this value is absolute or relative to another state depends on the implementing class.
        """
        self.level = level
        self.multiplicity = round(multiplicity) if multiplicity is not None else None
        # 'True' multiplicity is unrounded (do something smarter)
        self.true_multiplicity = multiplicity
        self.multiplicity_level = multiplicity_level
        self.energy = energy
        
    @property
    def multiplicity(self):
        """
        """
        return self._multiplicity if self._multiplicity is not None else 0
    
    @multiplicity.setter
    def multiplicity(self, value):
        """
        """
        self._multiplicity = value
    
    def __float__(self):
        """
        Float of this class.
        """
        return float(self.energy)
    
    @classmethod
    def multiplicity_number_to_string(self, multiplicity):
        if multiplicity == 1:
            return "singlet"
        elif multiplicity == 2:
            return "doublet"
        elif multiplicity == 3:
            return "triplet"
        elif multiplicity == 4:
            return"quartet"
        elif multiplicity is None or multiplicity == 0:
            return "no multiplicity"
        elif multiplicity % 1 == 0:
            # Multiplicity is an integer, so return as a stringy whole number.
            return str(int(multiplicity))
        else:
            return str(multiplicity)
        
    
    @property
    def multiplicity_string(self):
        return self.multiplicity_number_to_string(self.multiplicity)
    
    @property
    def multiplicity_symbol(self):
        multiplicity = self.multiplicity
        
        # Get a shorthand symbol if we can.
        if  multiplicity == 1:
            return "S"
        elif  multiplicity == 2:
            return "D"
        elif  multiplicity == 3:
            return "T"
        elif  multiplicity == 4:
            return "Q"
        elif multiplicity == 0:
            return "?"
        elif multiplicity % 1 == 0:
            # Multiplicity is an integer, so return as a stringy whole number.
            return str(int(multiplicity))
        else:
            return str(multiplicity)
    
    @property
    def state_symbol(self):
        """
        A short hand notation to identify this excited state.
        
        If the multiplicity is well defined (singlet, doublet, triplet etc), the symbol starts with an appropriate letter (S, D, T etc), otherwise a numeric multiplicity is used. The symbol ends with an integer in brackets, indicating the excited state's level.
        eg, S(1) is the first singlet excited state, S(2) is the second and so on.
        """
        return "{}({})".format(self.multiplicity_symbol, self.multiplicity_level)
    
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        return {
            "index": self.level,
            "symbol": self.state_symbol,
            "multiplicity": self.multiplicity,
            "multiplicity_index": self.multiplicity_level,
            "energy": {
                "value": float(self.energy),
                "units": "eV"
            }
        }


class Excited_state(Energy_state):
    """
    Class for representing an excited state.
    """
    
    def __init__(self, level, multiplicity, multiplicity_level, symmetry, energy, oscillator_strength, transitions, transition_dipole_moment = None):
        """
        Constructor for excited state objects.
        
        :param level: The 'level' of this excited state (essentially an index), where the lowest state has a level of 1, increasing by 1 for each higher state.
        :param multiplicity: The multiplicity of this excited state (as an integer or float).
        :param multiplicity_level: The 'level' of this state within the excited states that have the same multiplicity.
        :param symmetry: The symmetry of this excited state; a complicated term that contains our multiplicity among other things.
        :param energy: The energy of this excited state (in eV).
        :param oscillator_strength: The oscillator strength of this transition.
        :param transitions: The singly excited transitions which make up this transition.
        :param transition_dipole_moment: Optional transition dipole moment of the excited state.
        """
        super().__init__(level, multiplicity, multiplicity_level, energy)
        self.symmetry = symmetry
        # Oscillator strength is a dimensionless float that describes the probability of the transition from the reference state (normally ground) to this excited state.
        self.oscillator_strength = oscillator_strength
        # The transitions which contribute to this state.
        self.transitions = transitions
        
        # If we were given a TDM, set its excited state to ourself.
        if transition_dipole_moment is not None:
            transition_dipole_moment.set_excited_state(self)
                
        self.transition_dipole_moment = transition_dipole_moment
        
    @classmethod
    def get_multiplicity_from_symmetry(self, symmetry_string):
        """
        Get the multiplicity of of an excited state from its symmetry.
        
        See multiplicity() for a more detailed description of the format of multiplicity strings.
        :param symmetry_string: The symmetry string.
        :return: The multiplicity as a number (possibly an int, possibly a float).
        """
        # Split the symmetry string on the dash (-) character.
        multiplicity_string = (symmetry_string.split('-', 1))[0]
        # Convert to a number if we have a string.
        if multiplicity_string == "Singlet":
            return 1
        elif multiplicity_string == "Doublet":
            return 2
        elif multiplicity_string == "Triplet":
            return 3
        elif multiplicity_string == "Quartet":
            return 4
        elif multiplicity_string == "???":
            return None
        else:
            # Try and cast to float.
            # Float multiplicities don't make sense in the real world, but are possible results from calculations.
            return float(multiplicity_string)
    
    @property
    def wavelength(self):
        """
        The wavelength that corresponds to the energy of this excited state (in nm).
        
        :raises Result_unavailable_error: If the energy of this state is 0.
        """
        try:
        # λ = (c * h) / e
            return self.energy_to_wavelength(self.energy)
        except FloatingPointError:
            # Our energy is zero.
            raise Result_unavailable_error('excited state wavelength', "excited state '{}' energy is 0 eV".format(self.state_symbol))
        
    @property
    def joules(self):
        """
        The energy of this excited state in Joules.
        """
        return self.energy * 1.602176634e-19
        
    @property
    def color(self):
        """
        The 'color' that corresponds to the energy of this excited state (as a string).
        """
        # Go through our color presets and see which one we match.
        for color in self.colors:
            if self.wavelength < color["max"]:
                # Good match.
                return color["name"]
        return "???"
    
    @property
    def CIE_XYZ(self):
        """
        The CIE XYZ tristimulus values of the 'color' that corresponds to the energy of this excited state (as a numpy array).
        """
        import colour
        try:
            return colour.wavelength_to_XYZ(self.wavelength)
        except ValueError:
            # Wavelength is out of our colour range, so we can't see it.
            return numpy.zeros(3)
        
    @property
    def CIE_xy(self):
        """
        The CIE xy chromaticity coordinates of the 'color' that corresponds to the energy of this excited state.
        """
        import colour
        try:
            return colour.XYZ_to_xy(self.CIE_XYZ)
        except Exception:
            return numpy.zeros(2)
        
                
    @property
    def rgb(self):
        """
        The RGB values of the 'color' that corresponds to the energy of this excited state (as a list of [r, g, b] from 0 -> 255).
        """
        return self.xyz_to_rgb(self.CIE_XYZ)
        
    @classmethod
    def xyz_to_rgb(self, XYZ):
        import colour
        rgb =  [numpy.clip(clr, 0, math.inf) for clr in colour.XYZ_to_sRGB(XYZ)]
        
        # Now we normalise if one of our values exceeds 1.
        if max(rgb) > 1:
            rgb = [clr / max(rgb) for clr in rgb]
            
        # Now convert to 0 -> 255 and return.
        return [int(clr * 255) for clr in rgb]    
    
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        dump_dict = super()._dump_(digichem_options, all)
        dump_dict.update({
            "wavelength": {
                "value": float(self.wavelength),
                "units": "nm"
            },
            "colour": self.color,
            "cie": {
                "x": float(self.CIE_xy[0]),
                "y": float(self.CIE_xy[1]),
            },
            "symmetry": self.symmetry,
            "oscillator_strength": float(self.oscillator_strength) if self.oscillator_strength is not None else None,
            "tdm": self.transition_dipole_moment.dump(digichem_options, all) if self.transition_dipole_moment is not None else None,
            "transitions": [tran.dump(digichem_options, all) for tran in self.transitions],
        })
        return dump_dict
        
    @classmethod
    def list_from_parser(self, parser):
        """
        Create a list of Excited_state objects from an output file parser.
                
        :param parser: An output file parser.
        :return: A list of Excited_state objects in the same order as given in cclib.
        """
        # List of excited states.
        excited_states = []
        
        # Assemble cclib's various arrays into a single list.
        # If we're missing etoscs, that's ok.
        etoscsc = getattr(parser.data, "etoscs", [])
        
        try:
            excited_states_data = list(zip_longest(parser.data.etsyms, parser.data.etenergies, etoscsc, fillvalue = 0.0))
        
        except AttributeError:
            if not hasattr(parser.data, "etsyms") and not hasattr(parser.data, "etenergies"):
                return []
            
            elif not hasattr(parser.data, "etsyms"):
                digichem.log.get_logger().warning("Unable to fully parse excited states because only energies are available, missing symmetries")
                return []
            
            elif not hasattr(parser.data, "etenergies"):
                digichem.log.get_logger().warning("Unable to fully parse excited states because only symmetries are available, missing energies")
                return []
            
            else:
                raise
        
        # First get our transitions.
        transitions = Excited_state_transition.list_from_parser(parser)
        
        # Loop through our data.
        for index, (symmetry, energy, oscillator_strength) in enumerate(excited_states_data):                    
            # Relevant transition dipole moment.
            try:
                tdm = parser.results.transition_dipole_moments[index]
            except IndexError:
                tdm = None
                
            # Get and append our object.
            # We'll set level and mult level once we've got all our objects, so just use None for now.
            excited_states.append(
                self(
                    level = None,
                    multiplicity = Excited_state.get_multiplicity_from_symmetry(symmetry),
                    multiplicity_level = None,
                    symmetry = symmetry,
                    energy = self.wavenumbers_to_energy(energy),
                    oscillator_strength = oscillator_strength,
                    transitions = transitions[index] if len(transitions) != 0 else [],
                    transition_dipole_moment = tdm
                )
            )
        
        # Now assign total and multiplicity levels which we skipped earlier.
        Excited_state_list.assign_levels(excited_states)
        
        # All done, return our list.
        return excited_states
    
    @classmethod
    def list_from_dump(self, data, result_set, options):
        """
        Get a list of instances of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        states = []
        for es_dict in data:
            # Get our tdm (if available).
            if es_dict['tdm'] is not None:
                tdm = result_set.transition_dipole_moments[es_dict['index']-1]
            
            else:
                tdm = None
            
            states.append(self(
                es_dict['index'],
                es_dict['multiplicity'],
                es_dict['multiplicity_index'],
                es_dict['symmetry'],
                es_dict['energy']['value'],
                es_dict['oscillator_strength'],
                Excited_state_transition.list_from_dump(es_dict['transitions'], result_set, options),
                tdm
            ))
        
        return states
    


# General imports.
from itertools import zip_longest
import warnings

# Digichem imports.
from digichem.result import Result_container
from digichem.result import Result_object
from digichem.result import Floatable_mixin, Unmergeable_container_mixin

# Hidden.
#from digichem.result.spectroscopy import Spectroscopy_graph


class Vibrations_list(Result_container, Unmergeable_container_mixin):
    """
    Class for representing a group of molecular vibrations.
    """
    
    MERGE_WARNING = "Attempting to merge lists of vibrations that are not identical; non-equivalent vibrations will be ignored"
        
    @property
    def negative_frequencies(self):
        """
        Get a Vibrations_list object of the vibrations in this list that have negative frequencies (they are imaginary).
        """
        warnings.warn("negative_frequencies is deprecated, use negative instead")
        return self.negative
    
    @property
    def negative(self):
        """
        Get a Vibrations_list object of the vibrations in this list that have negative frequencies (they are imaginary).
        """
        return type(self)([vibration for vibration in self if vibration.frequency < 0])
    
    @classmethod
    def from_parser(self, parser):
        """
        Get an Vibrations_list object from an output file parser.
        
        :param parser: An output file parser.
        :return: A Vibrations_list object. The list will be empty if no vibration frequency data is available.
        """
        return self(Vibration.list_from_parser(parser))
    
    def _get_dump_(self):
        """
        Method used to get a dictionary used to generate on-demand values for dumping.
        
        This functionality is useful for hiding expense properties from the normal dump process, while still exposing them when specifically requested.
        
        Each key in the returned dict is the name of a dumpable item, each value is a function to call with digichem_options as its only param.
        """
        return {
            "spectrum": self.generate_spectrum,
        }
    
    def generate_spectrum(self, digichem_options):
        """
        Abstract function that is called to generate an on-demand value for dumping.
        
        This functionality is useful for hiding expense properties from the normal dump process, while still exposing them when specifically requested.
        
        :param key: The key being requested.
        :param digichem_options: Digichem options being used to dump.
        """
        from digichem.result.spectroscopy import Spectroscopy_graph
        
        
        spectrum = Spectroscopy_graph.from_vibrations(
            self,
            fwhm = digichem_options['IR_spectrum']['fwhm'],
            resolution = digichem_options['IR_spectrum']['gaussian_resolution'],
            cutoff = digichem_options['IR_spectrum']['gaussian_cutoff'],
            filter = digichem_options['IR_spectrum']['y_filter'],
        )
        
        
        try:
            spectrum_data = spectrum.plot_cumulative_gaussian()
            
            spectrum_data = [{"x":{"value": float(x), "units": "c m^-1"}, "y": {"value":float(y), "units": "km mol^-1"}} for x,y in spectrum_data]
            spectrum_peaks = [{"x":{"value": float(x), "units": "c m^-1"}, "y": {"value":float(y), "units": "km mol^-1"}} for x, y in spectrum.peaks()]
        
        except Exception:
            spectrum_data = []
            spectrum_peaks = []
        
        return {
            "values": spectrum_data,
            "peaks": spectrum_peaks
        }
    
    def _dump_(self, digichem_options, all):
        dump_dict = {
            "num_vibrations": len(self),
            "num_negative": len(self.negative),
            "values": super()._dump_(digichem_options, all),
        }
        return dump_dict

    @classmethod
    def from_dump(self, data, result_set, options):
        """
        Get an instance of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        return self(Vibration.list_from_dump(data['values'], result_set, options))


class Vibration(Result_object, Floatable_mixin):
    """
    Class for representing vibrational frequencies.
    """
    
    def __init__(self, level, frequency, intensity, symmetry = None):
        """
        Vibration object constructor.
        
        :param level: The numerical index of the vibration.
        :param frequency: Frequency of the vibration (cm-1).
        :param intensity: Intensity of the vibration in IR (km mol-1)
        :param symmetry: Symmetry term of the vibration (string).
        """
        self.level = level
        self.symmetry = symmetry
        self.frequency = frequency
        self.intensity = intensity
        
    def __float__(self):
        """
        A float representation of this object.
        """
        return float(self.frequency)
    
    def _dump_(self, digichem_options, all):
        """
        Get a representation of this result object in primitive format.
        """
        return {
            "index": self.level,
            "frequency": {
                "value": float(self.frequency),
                "units": "c m^-1",
            },
            "intensity": {
                "value": float(self.intensity),
                "units": "km mol^-1"
            },
            "symmetry": self.symmetry
        }
        
    @classmethod
    def list_from_dump(self, data, result_set, options):
        """
        Get a list of instances of this class from its dumped representation.
        
        :param data: The data to parse.
        :param result_set: The partially constructed result set which is being populated.
        """
        return [self(vib_dict['index'], vib_dict['frequency']['value'], vib_dict['intensity']['value'], vib_dict['symmetry']) for vib_dict in data]
        
    @classmethod
    def list_from_parser(self, parser):
        """
        Create a list of Vibration objects from an output file parser.
        
        :param parser: An output file parser.
        :return: A list of Vibration objects. The list will be empty if no frequency data is available.
        """
        try:
            return [self(index+1, frequency, intensity, symmetry) for index, (symmetry, frequency, intensity) in enumerate(zip_longest(getattr(parser.data, 'vibsyms', []), parser.data.vibfreqs, parser.data.vibirs, fillvalue = None))]
        except AttributeError:
            return []


"""This module contains translation tables between various different, but equivalent, terms expected by different calculation programs."""

import csv

from configurables.misc import is_int

import digichem.log
from digichem.datas import get_resource

# Hidden imports.
#import basis_set_exchange as bse


# Load functional conversion table from file.
functional_db = {}
with open(get_resource('data/functionals.csv')) as csv_file:
    _reader = csv.reader(csv_file)
    headers = None
    for line_num, line in enumerate(_reader):
        if line_num == 0:
            headers = line
        
        else:
            line = [line_part if line_part != "" else None for line_part in line]
            
            # The various parts of the definition.
            functional_name = line[0]
            aliases = line[1].split(",") if line[1] is not None else []
            functional = {"name": functional_name, "aliases": aliases, "turbomole": line[2], "gaussian": line[3]}
            
            # Get all the names we can refer to this functional by.
            all_names = [functional_name] + aliases
            for program in ("turbomole", "gaussian"):
                # Make sure we don't overwrite existing values with program specific ones.
                if functional[program] is not None and functional[program] not in functional_db:
                    all_names.append(functional[program])
            
            # Set into the DB, using all possible names.
            for ref_name in all_names:
                functional_db[ref_name] = functional


# Load solvent conversion table from file.
solvent_db = {}
with open(get_resource('data/solvents.csv')) as csv_file:
    _reader = csv.reader(csv_file)
    headers = None
    for line_num, line in enumerate(_reader):
        if line_num == 0:
            headers = line
        
        else:
            line = [line_part if line_part != "" else None for line_part in line]
            
            # The various parts of the definition.
            solvent = {"name": line[0], "aliases": [line[1]] if line[1] is not None else [], "gaussian": line[2], "epsilon": float(line[3]) if line[3] is not None else None, "refractive": float(line[4]) if line[4] is not None else None}
            
            solvent_db[solvent['name']] = solvent


class Translate():
    """
    ABC for translate classes.
    """
    
    def __init__(self, value):
        """
        Constructor for Translate classes.
        """
        if isinstance(value, type(self)):
            self.value = value.value
        
        else:
            self.value = value
    
    def translate(self, program):
        """
        Translate into a name appropriate for a given program.
        """
        raise NotImplementedError("Implement in subclass")
    
    def to_turbomole(self):
        """
        Translate into a name appropriate for Turbomole.
        """
        return self.translate("turbomole")
    
    def to_gaussian(self):
        """
        Translate into a name appropriate for Turbomole.
        """
        return self.translate("gaussian")
    
    def to_orca(self):
        """
        Translate into a name appropriate for ORCA.
        """
        return self.translate("orca")


class Basis_set(Translate):
    """
    A class for converting between the names of basis sets.
    """
    
    @property
    def basis_set(self):
        return self.value
    
    @classmethod
    def build_db(self):
        """
        Build a table of basis set names.
        
        :returns: A dictionary, where each key is the name of a basis set and each value is a BSE metadata dict.
        """
        import basis_set_exchange as bse
        # Build a database of basis set metadata, adding the display name and alternative names as additional keys for each basis set.
        # BSE's database has some semi-duplicate entries (such as '6-31g(d,p)' and '6-31g_st__st_').
        # Although these basis sets are identical, they have reversed metadata (ie, for '6-31g(d,p)', 'display_name' is set to '6-31G(d,p)' and 'other_names'
        # contains '6-31G**', while for '6-31g_st__st_' 'display_name' is '6-31G**' and 'other_names' is '6-31G(d,p)'.
        # This is a problem for us because we need to predictably choose from these names (ie the star form '6-31G**' is suitable for both
        # Gaussian and Turbomole), but the keys referring to these same names are different.
        #
        # Firstly, build a new dict using basename as the key, which appears to be the same even for duplicates.
        db = {value["basename"]: value for key, value in bse.get_metadata().items()}
        
        new_db = {}
        
        for basis_key, basis_set_meta in db.items():
            # First, add the normal key.
            basis_set = {
                "key": basis_set_meta['basename'],
                "name": basis_set_meta['display_name'],
                "aliases": basis_set_meta['other_names'],
                "meta": basis_set_meta
            }
            
            # Decide on our program specific names.
            basis_set["gaussian"] = basis_set['name']
            basis_set["turbomole"] = basis_set['name']
            
            # Turbomole doesn't like pople sets with polarisation function spelled out,
            # it prefers the star format.
            if basis_set['name'][-5:] == "(d,p)":
                basis_set["turbomole"] = basis_set['name'][:-5] + "**"
                
            # Gaussian has a strange, contracted style naming scheme for Karlsruhe,
            # and a misleading/incorrect name for def2-SV(P).
            #print(basis_set['name'])
            if basis_set['name'] == "def2-SV(P)":
                basis_set['gaussian'] = "def2SVPP"
                
            elif basis_set['name'][:5] == "def2-":
                basis_set['gaussian'] = basis_set['name'][:4] + basis_set['name'][5:]
            
            new_db[basis_key] = basis_set
            
            # TODO: Need to support aux basis sets (which end in things like -rifit, -jfit, -jkfit etc.
            # Orca, for example, uses /C, /J and /JK instead...
            # Turbomole uses nothing.
        
        # Basis set exchange seems to be missing some non-polarized Karlsruhe basis sets (or else Gaussian made them up).
        # Add them manually so we can still convert from Gaussian's weird representation.
        new_db.update({
            "def2-TZV": {"key": "def2-TZV", "name": "def2-TZV", "aliases": [], "meta": {}, "gaussian": "def2TZV"},#, "turbomole": "def2-TZV"},
            "def2-QZV": {"key": "def2-QZV", "name": "def2-QZV", "aliases": [], "meta": {}, "gaussian": "def2QZV"}#, "turbomole": "def2-TZV"}
        })
                
        return new_db
    
    @classmethod
    def find_in_db(self, hint):
        """
        Try and find the entry for a basis set in the basis set exchange.
        
        :raises ValueError: If the given hint could not be found.
        :param hint: The name of the basis set to search for.
        :return: A dictionary of BSE metadata.
        """
        # Get the basis set DB.
        db = self.build_db()
        
        # First, try and see if we can just use hint as an exact match.
        try:
            return db[hint]
            
        except KeyError:
            pass
        
        # Try again, this time ignoring case and looking through all possible names.
        str_hint = "".join(str(hint).lower().split())
        for basis_set in db.values():
            names = [basis_set.get(name, "") for name in ("key", "name", "gaussian", "turbomole")] + basis_set['aliases']
            if str_hint in [name.lower() for name in names if name is not None]:
                return basis_set
        
        # No luck.
        raise ValueError("Could not find basis set definition for '{}'".format(hint))
    
    def translate(self, program = "name"):
        """
        Translate the name of this basis set into one appropriate for a calculation program.
        
        Basis sets are currently handled the same for all programs.
        """
        try:
            basis_def = self.find_in_db(self.basis_set)
            return basis_def[program]
        
        except ValueError:
            # Just return as is.
            if self.basis_set != "auto":
                digichem.log.get_logger().debug("Could not find basis set with name '{}' in the basis set exchange; using name unmodified".format(self.basis_set))
            return self.basis_set
    
        except KeyError:
            return basis_def['name']
    
    def __str__(self):
        return str(self.translate())


class Functional(Translate):
    """
    A class for converting between the names of DFT functionals.
    """
    
    @property
    def functional(self):
        return self.value
            
    
    @classmethod
    def find_in_db(self, hint):
        """
        Try and find an entry for a functional in the internal library.
        
        :raises ValueError: If the functional could not be found.
        :param hint: The name of the functional to look for.
        :returns: The corresponding functional dict.
        """
        # First try using exact name.
        try:
            return functional_db[hint]
            
        except KeyError:
            pass
        
        # Try again, this time ignoring case.
        functional_names = [functional_name.upper() for functional_name in functional_db]
        
        try:
            return functional_db[list(functional_db.keys())[functional_names.index(hint.upper())]]
        
        except IndexError:
            pass
        
        # No luck.
        raise ValueError("Could not find functional definition for  '{}'".format(hint))
        
    
    def translate(self, program):
        """
        Translate this functional name to one recognised by a given program.
        
        :param program: The program to translate for.
        """
        # Try and get a definition for the functional from our db.
        try:
            func_def = self.find_in_db(self.functional)
                
            # If there's an explicit value for our program, use that.
            if program in func_def and func_def[program] is not None:
                return func_def[program]
            
            # Otherwise, use the main common name.
            return func_def['name']
            
        except ValueError:
            pass
        
        # No definition found, use as is.
        return self.functional
    
    def to_turbomole(self):
        """
        Translate into a name appropriate for Turbomole.
        """
        func_name = self.translate("turbomole")
        # All turbomole functionals are lower case, except for the word 'Gaussian'.
        return func_name.lower().replace("gaussian", "Gaussian")
    
    def __str__(self):
        return str(self.translate("name"))


class Solvent(Translate):
    """A class for converting between different representations of solvents."""
    
    @property
    def solvent(self):
        return self.value
    
    @classmethod
    def find_in_db(self, hint):
        """
        Try and find an entry for a solvent in the internal library.
        
        :raises ValueError: If the multiplicity could not be found.
        :param hint: The name, number or symbol of the multiplicity to look for.
        :returns: The corresponding multiplicity dict.
        """
        # First, just try to lookup exactly.
        try:
            return solvent_db[hint]
            
        except KeyError:
            # No luck, look through for a match.
            str_hint = "".join(str(hint).lower().split())
            for solvent in solvent_db.values():
                if hint == solvent['epsilon'] or str_hint in [name.lower() for name in solvent['aliases'] + [solvent['name']] + [solvent['gaussian']] if name is not None]:
                    return solvent
        
        # No luck.
        raise ValueError("Could not find solvent definition for '{}'".format(hint))
    
    @classmethod
    def epsilon_to_name(self, epsilon, threshold = 0.0001):
        """
        Find the name of a solvent based on its epsilon value.
        """
        close = {}
        for solvent_def in solvent_db.values():
            if solvent_def['epsilon'] == epsilon:
                # Exact match, stop.
                return solvent_def['name']
            
            elif abs(solvent_def['epsilon'] - epsilon) < threshold:
                # Close match.
                close[abs(solvent_def['epsilon'] - epsilon)] = solvent_def['name']
                
        # Sort values that were close based on how close they were.
        close_names = [close[closeness] for closeness in sorted(close)]
        
        try:
            return close_names[0]
        
        except IndexError:
            raise ValueError("No solvent definitions within {} of {}".format(threshold, epsilon)) from None
    
    @property
    def epsilon(self):
        return self.find_in_db(self.solvent)['epsilon']
    
    @property
    def refractive_index(self):
        return self.find_in_db(self.solvent)['refractive']
    
    def translate(self, program):
        """
        Translate this solvent name to one recognised by a given program.
        
        :param program: The program to translate for.
        """
        if program == "turbomole":
            program = "epsilon"
            
        elif program == "orca":
            program = "name"
        
        try:
            solvent_def = self.find_in_db(self.solvent)
            
            if program in solvent_def and solvent_def[program] is not None:
                return solvent_def[program]
            
            else:
                return solvent_def["name"]
        
        except ValueError:
            return self.solvent
        
    def __str__(self):
        return str(self.translate("name"))


class SCF_convergence(Translate):
    """
    A class for converting between different shortcuts for SCF convergence thresholds.
    """
    
    table = [
        {"name": "Loose",       "energy": -5,       "density": -3,      "orca": "LooseSCF"},   # Note energy stops changing here.
        {"name": "Weak",        "energy": -5,       "density": -4,      "orca": "SloppySCF"},                     
        {"name": "Medium",      "energy": -6,       "density": -5,      "orca": "MediumSCF"},
        {"name": "Strong",      "energy": -7,       "density": -6,      "orca": "StrongSCF"},
        {"name": "Tight",       "energy": -8,       "density": -7,      "orca": "TightSCF"},   # Typically a sensible default.
        {"name": "VTight",      "energy": -9,       "density": -8,      "orca": "VeryTightSCF"},
        {"name": "VVTight",     "energy": -10,      "density": -9,      "orca": "VeryVeryTightSCF"},
        {"name": "Extreme",     "energy": -14,      "density": -14,     "orca": "ExtremeSCF"},
    ]
    
    @classmethod
    def names(self):
        return [row['name'] for row in self.table]
    
    @classmethod
    def choices(self):
        return [self(row['name']) for row in self.table]

    @classmethod
    def find_in_db(self, hint):
        """
        Try and find an entry in the internal library.
        
        :raises ValueError: If the value could not be found.
        :param hint: The name to look for
        :returns: The corresponding value dict.
        """
        for row in self.table:
            name, energy, density, orca = row.values()
            
            if str(hint).upper() == name.upper() or str(hint).upper() == orca.upper():
                return row
            
            if str(hint) == str(energy) or str(hint).upper() == "ENERGY{}".format(energy) or str(hint).upper() == "DENSITY{}".format(density):
                return row
            
        # No luck.
        raise ValueError("Could not find convergence definition for '{}'".format(hint))
    
    def translate(self, to_type):
        """
        Translate into a name appropriate for a given program.
        """
        try:
            return self.find_in_db(self.value)[to_type]
        
        except ValueError:
            # Couldn't find in conversion table.
            return self.value
        
    def __str__(self):
        return str(self.translate("name"))
    
    def __eq__(self, other):
        return str(self) == str(other)
    

class Cube_grid_points(Translate):
    """A class for converting between cube grid sizes."""
    
    table = [
        {"name": "Default",     "points": 100,      "gaussian": 0,       "turbomole": "m3",      "orca": 100},   # Default depends on the calc program. Gaussian uses a special value of 'zero' to select a default algorithm.
        {"name": "Tiny",        "points": 25,                            "turbomole": "1",                  },
        {"name": "Small",       "points": 50,                            "turbomole": "2",                  },
        {"name": "Medium",      "points": 100,                           "turbomole": "3",                  },
        {"name": "Large",       "points": 200,                           "turbomole": "4",                  },
        {"name": "Huge",        "points": 500,                           "turbomole": "5",                  },
        
    ]

    @classmethod
    def find_in_db(self, hint):
        """
        Try and find an entry in the internal library.
        
        :raises ValueError: If the value could not be found.
        :param hint: The name to look for
        :returns: The corresponding value dict.
        """
        for row in self.table:
            if str(hint).upper() == row['name'].upper():
                return row
            
        # No luck.
        raise ValueError("Could not find grid point definition for '{}'".format(hint))
    
    def translate(self, to_type):
        """
        Translate into a name appropriate for a given program.
        """
        try:
            grid_def = self.find_in_db(self.value)
            return grid_def[to_type]
        
        except KeyError:
            # Just return the equivalent number of points.
            return grid_def['points']
        
        except ValueError:
            # Couldn't find in conversion table.
            return self.value
        
    def __str__(self):
        return str(self.translate("name"))


# TODO: Use this.
class Multiplicity(Translate):
    """A class for converting between different representations of multiplicity."""
    
    table = [
        {"name": "no multiplicity", "number": 0, "symbol": "?"},
        {"name": "Singlet", "number": 1, "symbol": "S"},
        {"name": "Doublet", "number": 2, "symbol": "D"},
        {"name": "Triplet", "number": 3, "symbol": "T"},
        {"name": "Quartet", "number": 4, "symbol": "Q"}
    ]
    
    @property
    def multiplicity(self):
        return self.value
        
    @classmethod
    def find_in_db(self, hint):
        """
        Try and find an entry for a multiplicity in the internal library.
        
        :raises ValueError: If the multiplicity could not be found.
        :param hint: The name, number or symbol of the multiplicity to look for.
        :returns: The corresponding multiplicity dict.
        """
        for row in self.table:
            name, number, symbol = row.values()
            if hint == number or str(hint).upper() in (str(number), name.upper(), symbol.upper()):
                return row
            
        # No luck.
        raise ValueError("Could not find multiplicity definition for '{}'".format(hint))
    
    @property
    def symbol(self):
        """
        Get this multiplicity as a symbol.
        """
        # Get a shorthand symbol if we can.
        try:
            return self.find_in_db(self.multiplicity)['symbol']
            
        except ValueError:
            if self.multiplicity % 1 == 0:
                # Multiplicity is an integer, so return as a stringy whole number.
                return str(int(self.multiplicity))
            else:
                return str(self.multiplicity)
            
    @property
    def string(self):
        """
        Get this multiplicity as a string.
        """
        return self.translate("name")
    
    def __str__(self):
        return self.string
    
    @property
    def number(self):
        """
        Get this multiplicity as a number.
        """
        try:
            return self.find_in_db(self.multiplicity)['number']
        
        except ValueError:
            # No pre-defined number, see if it is an int.
            if is_int(self.multiplicity):
                return int(self.multiplicity)
            
            else:
                return float(self.multiplicity)
    
    def translate(self, to_type):
        """
        Translate into a name appropriate for a given program.
        """
        if to_type == "gaussian":
            to_type = "name"
        
        elif to_type == "turbomole":
            to_type = "number"
        
        
        try:
            return self.find_in_db(self.multiplicity)[to_type]
        
        except ValueError:
            # Couldn't find in conversion table.
            return self.multiplicity
